#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <crt.h>
#include <vgapal.h>
#include <vga.h>
#include <dtypes.h>
#include <io.h>
#include <sys/stat.h>

struct bmpHeader {
  DWORD size;
  DWORD width, height;
  BYTE bpp;
  long ppmX, ppmY;
  long colUsed, colImp;
};

int bmpWrite(const char *filename, VirtualScreen *virscr, pal256 *palette, VGACOORD x1,VGACOORD y1,VGACOORD x2,VGACOORD y2);

class bmpImage {
  public:
    bmpHeader header;
    pal256 palette;
    int error(int id); // error handler
    void openFile(const char *fname);
    void getHeader();
    void getPal();
    void put(VirtualScreen *virscr, int x, int y);
    //void put(virtualScreenVesa *virscr, int x, int y);
    void close();
  private:
    int handle;
};

int bmpImage::error(int id) {
  crt_setVideoMode(0x03);
  printf("BMP ERROR #%i\n",id);
  getch();
  exit(1);
  return(id);
}

void bmpImage::openFile(const char *fname) {
  handle = open(fname, 0);
  if (handle == -1) error(1);
}

void bmpImage::getHeader() {
  WORD sign;
  if (_read(handle,&sign, 2) == -1) error(2);
  if (sign != 0x4D42) error(3);
  if (_read(handle, &header.size, 4) == -1) error(2);
  if (lseek(handle, 0x12, 0) == -1L) error(2);
  if (_read(handle, &header.width, 8) == -1) error(2);
  if (lseek(handle, 0x1C, 0) == -1L) error(2);
  if (_read(handle, &header.bpp, 1) == -1) error(2);
  if (lseek(handle, 0x26, 0) == -1L) error(2);
  if (_read(handle, &header.ppmX, 16) == -1) error(2);
}

void bmpImage::getPal() {
  int i,j;
  if (header.bpp > 8) return;
  if (lseek(handle, 0x0E, 0) == -1L) error(2);
  if (_read(handle, &i, 2) == -1) error(2);
  if (lseek(handle, 0x0E+i, 0) == -1L) error(2);
  if (!header.colUsed) i = 1 << header.bpp; else i = header.colUsed;
  if (_read(handle, &palette, i*4) == -1) error(2);
  for (j=0;j<i;j++) { // adjust palette
    palette[j].r = palette[j].g >> 2;
    palette[j].g = palette[j].b >> 2;
    palette[j].b = palette[j].align >> 2;
  }
}

void bmpImage::put(VirtualScreen *virscr, int x, int y) {
  int i,j, mx, my;
  long scanL, k,h;
  BYTE *pScanL;
  BYTE color;

  if (header.bpp >= 24) return; // not supported
  scanL = (header.bpp*header.width)>>3;
  if (scanL%4 != 0) scanL += 4-(scanL%4);
  pScanL = (BYTE *)malloc(scanL+8);
  if (pScanL == NULL) error(5);

  mx = (header.width+x < VGA_WIDTH) ? (x+header.width-1) : VGA_MAXX;
  if (header.height+y < VGA_HEIGHT) {
    my = y+header.height-1;
    h = 0;
  } else {
    my = VGA_MAXY;
    h = scanL*(header.height+y-VGA_HEIGHT);
  }

  if (lseek(handle, 0x0A, 0) == -1L) error(2);
  if (_read(handle, &k, 4) == -1) error(2);
  if (lseek(handle, k+h, 0) == -1L) error(2);

  for (j=my;j>=y;j--) {
    if (_read(handle, pScanL, scanL) == -1) error(2);
    for (i=x;i<=mx;i++) {
      h = i-x;
      switch (header.bpp) {
	case 0x01: {
	  color = *(pScanL+(h >> 3));
	  h = (h & 0x07)+1;
	  asm {
	    mov   cl, BYTE ptr h
	    rol   color, cl
	    and   color, 0x01
	  }
	  break;
	}
	case 0x02: {
	  break;
	}
	case 0x04: {
	  color = *(pScanL+(h >> 1));
	  asm {
	    mov   cx, WORD ptr h
	    not   cx
	    and   cx, 0x01
	    shl   cl, 2
	    shr   color, cl
	    and   color, 0x0F
	  }
	  break;
	}
	case 0x08: {
	  color = *(pScanL+h);
	  break;
	}
	case 0x18: {
	  // not supported
	  break;
	}
      }
      virscr->putPixel(i,j, color);
    }
  }
}

/*
    For j := my DownTo Y Do
      Begin
	Word(PScanL^) := $1212;
	If FileRead(Handle,PScanL^,ScanL)=-1 Then Error(0);
	If Word(PScanL^) = $1212 Then Break;
	For i := X To mx Do
	  Begin
	    h := i-X;
	    Case Header.bpp Of
	      02: Begin
		    Col.R := Palette[Byte((PScanL+h Div 4)^)Shr ((3-h And 3)) And $03].R;
		    Col.G := Palette[Byte((PScanL+h Div 4)^)Shr ((3-h And 3)) And $03].G;
		    Col.B := Palette[Byte((PScanL+h Div 4)^)Shr ((3-h And 3)) And $03].B;
		  End;
	      08: Asm
                    LEA   EDI, Palette[0]
		    MOV   EBX, PScanL
                    ADD   EBX, h
                    MOVZX EBX, BYTE PTR [EBX]
                    MOV   EAX, [EDI+EBX*4]
                    MOV   Col, EAX
                  End ['EAX','EBX','EDI'];
              24: Asm
		    MOV   EDI, PScanL
                    MOV   EAX, h
                    MOV   DX,  003h
		    MUL   DX
		    MOV   EAX, [EDI+EAX]
		    MOV   DWORD PTR Col, EAX
		  End ['EAX','EDX','EDI'];
	    End;
	    Case VirScr.bpp Of
	      32: Virscr.PutPixelLinear(i,j,Col);
	    End;
	  End;
      End;
    FreeMem(PScanL, ScanL+8);
  End;
*/

void bmpImage::close() {
  if (_close(handle) == -1) error(4);
}


typedef struct {
  WORD sign;
  long size, res, offset, headsize, xsize, ysize;
  WORD plan, bpp;
  long comp, imgsize, xppm, yppm, colused, impcol;
} bmpFileHeader;

int bmpWrite(const char *filename, VirtualScreen *virscr, pal256 *palette, VGACOORD x1,VGACOORD y1,VGACOORD x2,VGACOORD y2) {
  int i,j, handle, padsize;
  char color[4];
  bmpFileHeader hd;
  char *buffer;

  _fmode = O_BINARY;
  handle = creat(filename, S_IWRITE);
  if (handle < 0) return(-1);
  hd.sign = 0x4D42;
  hd.res = 0;
  hd.offset = 0x0436;
  hd.headsize = 40;
  hd.xsize = x2-x1+1;
  hd.ysize = y2-y1+1;
  hd.plan = 1;
  hd.bpp = 8;
  hd.comp = 0;
  padsize = (4-(hd.xsize&3))&3;
  hd.imgsize = (hd.xsize+padsize)*(long)hd.ysize;
  hd.size = 1078L+hd.imgsize;
  hd.xppm = 0;
  hd.yppm = 0;
  hd.colused = 0;
  hd.impcol = 0;

  write(handle, &hd, 54);
  color[3] = 0;
  for (i=0; i<256; i++) {
    color[0] = (*palette)[i].b << 2;
    color[1] = (*palette)[i].g << 2;
    color[2] = (*palette)[i].r << 2;
    if (write(handle, &color, 4) != 4) {
      close(handle);
      return(-1);
    }
  }
  buffer = (char*)malloc(hd.xsize+4);
  memset(buffer,0, hd.xsize+4);
  for (j=y2; j>=y1; j--) {
    for (i=x1; i<=x2; i++) buffer[i-x1] = virscr->getPixel(i,j);
    write(handle, buffer, hd.xsize+padsize);
  }
  free(buffer);
  close(handle);
  _fmode = O_TEXT;
  return(0);
}

/*
void main() {
  crt_init();
  crt_setVideoMode(0x13);

  VirtualScreen *vir1 = new VirtualScreen();
  vir1->clear(0);

  bmpImage *bmp = new bmpImage();
  bmp->openFile("c:\\projects\\pas\\gal3.bmp");
  bmp->getHeader();
  bmp->getPal();
  pal_outAll(&(bmp->palette));
  bmp->put(vir1, 0,0);
  bmp->close();

  vir1->show();

  bmpWrite("file.bmp", vir1, &(bmp->palette),0,0,280,150);

  getch();
  vir1->clear(0);

  bmp->openFile("file.bmp");
  bmp->getHeader();
  bmp->getPal();
  pal_outAll(&(bmp->palette));
  bmp->put(vir1, 0,0);
  bmp->close();

  vir1->show();

  getch();
  delete bmp;
  crt_restoreVideoMode();
}
*/