/*
CRT.H

A collection of functions to manipulate text-mode screen
*/

#ifndef CRT_H
#define CRT_H

#include <dtypes.h>

char crt_checkCoord = 0; // if enabled crt_print funcs check input coords

const WORD crt_videoSeg = 0xB800;
WORD crt_columns = 80;
WORD crt_lines = 25;
BYTE crt_activeMode = 3;
BYTE crt_lastMode = 3;
BYTE crt_activePage = 0;
WORD crt_pageSize = 4096;


#pragma warn -rvl
BYTE crt_getVideoMode(void) {
  asm {
    push es
    mov  ax, 0x0040
    mov  es, ax
    mov  al, es:[0x0049]
    mov  crt_activeMode, al
    pop  es
  }
}
#pragma warn +rvl

void crt_setVideoMode(BYTE mode) {
  crt_lastMode = crt_getVideoMode();
  asm {
    xor  ax,ax
    mov  al, mode
    int  0x10
  }
  crt_getVideoMode();
}

void crt_restoreVideoMode() {
  asm {
    xor  ax,ax
    mov  al, crt_lastMode
    int  0x10
  }
  crt_getVideoMode();
}

#pragma warn -rvl
BYTE crt_getColumns(void) {
  asm {
    push es
    mov  ax, 0x0040
    mov  es, ax
    mov  al, es:[0x004A]
    mov  byte ptr crt_columns, al
    pop  es
  }
}
#pragma warn +rvl

#pragma warn -rvl
BYTE crt_getPage(void) {
  asm {
    push es
    mov  ax, 0x0040
    mov  es, ax
    mov  al, es:[0x0062]
    mov  crt_activePage, al
    pop  es
  }
}
#pragma warn +rvl

void crt_setPage(BYTE page) {
  asm {
    mov  ah, 0x05
    mov  al, page
    int  0x10
  }
  crt_getPage();
}

#pragma warn -rvl
WORD crt_getPageSize(void) {
  asm {
    push es
    mov  ax, 0x0040
    mov  es, ax
    mov  ax, es:[0x004C]
    mov  crt_pageSize, ax
    pop  es
  }
}
#pragma warn +rvl

void crt_enableBlink(char enable) { // 1=yes, 0=no
  asm {
    push bx
    mov  ax, 0x1003
    mov  bl, enable
    int  0x10
    pop  bx
  }
}

void crt_setCursorPos(short x, short y) {
  asm {
    push bx
    push dx
    mov  bh, crt_activePage
    mov  dl, byte ptr x
    mov  dh, byte ptr y
    mov  ah, 0x02
    int  0x10
    pop  dx
    pop  bx
  }
}

void crt_cls() {
  asm {
    push es
    push di
    push cx
    push dx
    mov  ax, crt_videoSeg
    mov  es, ax
    mov  al, crt_activePage
    xor  ah, ah
    mul  crt_pageSize
    mov  di, ax
    mov  ax, crt_columns
    mul  crt_lines
    mov  cx, ax
    mov  ax, 0x0720
    rep  stosw
    mov  bh, crt_activePage
    mov  ah, 0x02
    xor  dx, dx
    int  0x10   // set cursor
    pop  dx
    pop  cx
    pop  di
    pop  es
  }
}

void crt_printc(short x, short y, BYTE attr, char text) {
  asm {
    push es
    push bx
    push dx
    cmp  crt_checkCoord, 0
    je   dontCheck
    mov  ax, x
    cmp  ax, 0
    jl   endf
    cmp  ax, crt_columns
    jge  endf
    mov  ax, y
    cmp  ax, 0
    jl   endf
    cmp  ax, crt_lines
    jge  endf
  } dontCheck: asm {
    mov  ax, crt_videoSeg
    mov  es, ax
    mov  al, crt_activePage
    xor  ah, ah
    mul  crt_pageSize
    mov  bx, ax
    mov  ax, y
    mul  word ptr crt_columns
    add  ax, x
    shl  ax, 1
    add  bx, ax
    mov  dl, text
    mov  dh, attr
    mov  es:[bx], dx
   } endf: asm {
    pop  dx
    pop  bx
    pop  es
  }
}

#pragma warn -rvl
short crt_getc(short x, short y) {
  asm {
    push es
    push bx
    push dx
    cmp  crt_checkCoord, 0
    je   dontCheck
    mov  ax, x
    cmp  ax, 0
    jl   endf
    cmp  ax, crt_columns
    jge  endf
    mov  ax, y
    cmp  ax, 0
    jl   endf
    cmp  ax, crt_lines
    jge  endf
  } dontCheck: asm {
    mov  ax, crt_videoSeg
    mov  es, ax
    mov  al, crt_activePage
    xor  ah, ah
    mul  crt_pageSize
    mov  bx, ax
    mov  ax, y
    mul  word ptr crt_columns
    add  ax, x
    shl  ax, 1
    add  bx, ax
    mov  ax, es:[bx]
   } endf: asm {
    pop  dx
    pop  bx
    pop  es
  }
}
#pragma warn +rvl

void crt_print(short x, short y, BYTE attr, char *text) {
  asm {
    push es
    push di
    push bx
    push dx
    cmp  crt_checkCoord, 0
    je   dontCheck
    mov  ax, y
    cmp  ax, 0
    jl   endf
    cmp  ax, crt_lines
    jge  endf
  } dontCheck: asm {
    mov  ax, crt_videoSeg
    mov  es, ax
    mov  al, crt_activePage
    xor  ah, ah
    mul  crt_pageSize
    mov  di, ax
    mov  ax, y
    mul  word ptr crt_columns
    add  ax, x
    shl  ax, 1
    add  di, ax
    mov  ah, attr
    mov  bx, text
    mov  dx, x
  } beginloop: asm {
    mov  al, byte ptr [bx]
    cmp  al, 0
    je   endloop
    cmp  crt_checkCoord, 0
    je   dontCheck2
    cmp  dx, 0
    jnl  continueloop
    add  di, 2
    inc  bx
    inc  dx
    jmp  beginloop
  } continueloop: asm {
    cmp  dx, crt_columns
    jge  endf
  } dontCheck2: asm {
    stosw
    inc  bx
    inc  dx
    jmp  beginloop
  } endloop: asm {
  } endf: asm {
    pop  dx
    pop  bx
    pop  di
    pop  es
  }
}

void crt_copyPage(BYTE dest, BYTE src) {
  asm {
    push ds
    push es
    push di
    push si
    push cx
    push dx
    mov  ax, crt_videoSeg
    mov  es, ax
    mov  cx, crt_pageSize
    mov  ds, ax
    xor  ax, ax
    mov  al, src
    mul  cx
    mov  si, ax
    xor  ax, ax
    mov  al, dest
    mul  cx
    mov  di, ax
    shr  cx, 1
    rep  movsw
    pop  dx
    pop  cx
    pop  si
    pop  di
    pop  es
    pop  ds
  }
}

void crt_init() {
  crt_getVideoMode();
  crt_getColumns();
  crt_getPageSize();
  crt_getPage();
}

#endif