/*
   VESA
*/

#ifndef VESA
#define VESA

#pragma option -2
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>

const supported = 0x4F;
const t80x60         = 0x0108,    t132x25        = 0x0109,  // Video modes
      t132x43        = 0x010A,    t132x50        = 0x010B,  //
      t132x60        = 0x010C,                              //
      g640x400x256   = 0x0100,    g640x480x256   = 0x0101,  //
      g800x600x16    = 0x0102,    g800x600x256   = 0x0103,  //
      g1024x768x16   = 0x0104,    g1024x768x256  = 0x0105,  //
      g1280x1024x16  = 0x0106,    g1280x1024x256 = 0x0107,  //
      g320x200x32K   = 0x010D,    g320x200x64K   = 0x010E,  //
      g320x200x16M   = 0x010F,    g640x480x32K   = 0x0110,  //
      g640x480x64K   = 0x0111,    g640x480x16M   = 0x0112,  //
      g800x600x32K   = 0x0113,    g800x600x64K   = 0x0114,  //
      g800x600x16M   = 0x0115,    g1024x768x32K  = 0x0116,  //
      g1024x768x64K  = 0x0117,    g1024x768x16M  = 0x0118,  //
      g1280x1024x32K = 0x0119,    g1280x1024x64K = 0x011A,  //
      g1280x1024x16M = 0x011B,    g1600x1200x256 = 0x0120,  //
      g1600x1200x32K = 0x0121,    g1600x1200x64K = 0x0122,  //
      gSpecial       = 0x81FF;                              //
const UseCRTCRefresh  = 0x0800,  // Constants to add to video modes
      UseLinearBuffer = 0x4000,  //
      PreserveMemory  = 0x8000;  //
const mmText     = 0x00,   mmCGA      = 0x01,  // Constants for MemoryModel
      mmHercules = 0x02,   mmPlanar   = 0x03,  // in ModeInfoBlock structure
      mmPacked   = 0x04,   mmNonChain = 0x05,  //
      mmDirect   = 0x06,   mmYUV      = 0x07;  //

char * TF[2] = {"FALSE","TRUE"};
char * AP[2] = {"ABSENT","PRESENT"};
char * mm[8] = {"Text mode","CGA","Hercules","Planar",
		"Packed pixel","Non-chain 4","Direct color","YUV"};


char VbeStatus;

struct VbeInfoBlock {
  char       VbeSignature[4];
  union {
    int        VbeVersion;
    struct {
      unsigned Low:8;
      unsigned High:8;
    } VbeVersionF;
  };
  char far * OemStringPtr;
  union {
    char Capabilities[4];
    struct {
      unsigned SwitchableDAC:1;
      unsigned VGACompatible:1;
      unsigned UseBlankBit:1;
      unsigned StereoscopicSupported:1;
      unsigned StereoViaVESAEVC:1;
    } CapabilitiesF;
  };
  void far * VideoModePtr;
  int        TotalMemory;
  int        OemSoftwareRev;
  char far * OemVendorNamePtr;
  char far * OemProductNamePtr;
  void far * OemProductRevPtr;
  char       Reserved[222];
  char       OemData[256];
};

typedef struct {
	  unsigned Relocatable:1;
	  unsigned Readable:1;
	  unsigned Writeable:1;
	} WinAttributes;

struct ModeInfoBlock {
  union {                          // All VBE version
    int ModeAttributes;
    struct {
      unsigned Supported:1;
      unsigned Reserved:1;
      unsigned TTYOutSupported:1;
      unsigned ColorMode:1;
      unsigned GraphicsMode:1;
      unsigned VGACompatible:1;
      unsigned VGACompatibleWinMode:1;
      unsigned LinearBuffer:1;
      unsigned DoubleScan:1;
      unsigned Interlaced:1;
      unsigned TripleBuffering:1;
      unsigned StereoscopicSupported:1;
      unsigned DualDisplayStart:1;
      unsigned Reserved2:3;
    } ModeAttributesF;
  };
  union {
    char    WinAAttributes;
    WinAttributes WinAAttributesF;
  };
  union {
    char    WinBAttributes;
    WinAttributes WinBAttributesF;
  };
  int        WinGranularity;
  int        WinSize;
  int        WinASegment;
  int        WinBSegment;
  void far * WinFuncPtr;
  int        BytesPerScanLine;
  int        XResolution;          // VBE 1.2+
  int        YResolution;
  char       XCharSize;
  char       YCharSize;
  char       NumberOfPlanes;
  char       BitsPerPixel;
  char       NumberOfBanks;
  char       MemoryModel;
  char       BankSize;
  char       NumberOfImagesPages;
  char       Reserved;
  char       RedMaskSize;
  char       RedFieldPosition;
  char       GreenMaskSize;
  char       GreenFieldPosition;
  char       BlueMaskSize;
  char       BlueFieldPosition;
  char       RsvdMaskSize;
  char       RsvdFieldPosition;
  char       DirectColorModeInfo;
  void far * PhysBasePtr;          // VBE 2.0+
  long       Reserved2;
  int        Reserved3;
  int        LinBytesPerScanLine;  // VBE 3.0+
  char       BnkNumberOfImagePages;
  char       LinNumberOfImagePages;
  char       LinRedMaskSize;
  char       LinRedFieldPosition;
  char       LinGreenMaskSize;
  char       LinGreenFieldPosition;
  char       LinBlueMaskSize;
  char       LinBlueFieldPosition;
  char       LinRsvdMaskSize;
  char       LinRsvdFieldPosition;
  long       MaxPixelClock;
  char       Reserved4[189];
};

struct CRTCInfoBlock {
  int  HorizontalTotal;
  int  HorizontalSyncStart;
  int  HorizontalSyncEnd;
  int  VerticalTotal;
  int  VerticalSyncStart;
  int  VerticalSyncEnd;
  union {
    char Flags;
    struct {
      unsigned DoubleScanned:1;
      unsigned Interlaced:1;
      unsigned Hpolarityneg:1;
      unsigned Vpolarityneg:1;
    } FlagsF;
  };
  long PixelClock;
  int  RefreshRate;
  char Reserved[40];
};

void VbeErrorHandler(char supp) {
  if (supp == supported) return;
  asm {
    MOV  AX, 0x0003
    INT  0x10
  }
  printf("VESA error %Xh\nPress a key to exit...",VbeStatus);
  getch();
  exit(1);
};

void (*VbeErrorPtr)(char supp) = &VbeErrorHandler;
VbeInfoBlock vbeinfo;
ModeInfoBlock modeinfo;
CRTCInfoBlock crtinfo;
VbeInfoBlock far * vbeinfoPtr = &vbeinfo;
ModeInfoBlock far * modeinfoPtr = &modeinfo;
CRTCInfoBlock far * crtinfoPtr = &crtinfo;


#pragma option -w-rvl
char VbeGetInfo(void) {
  asm {
    PUSH DI; PUSH ES
    MOV  AX, 0x4F00
    LES  DI, vbeinfoPtr
    INT  0x10
    MOV  VbeStatus, AH
    POP  ES; POP  DI
    CMP  AX, 0x004F
    JE   ok
  }
  VbeErrorPtr(_AL);
  ok:
};

char VbeGetModeInfo(int mode) {
  asm {
    PUSH DI; PUSH ES
    MOV  AX, 0x4F01
    MOV  CX, mode
    LES  DI, modeinfoPtr
    INT  0x10
    MOV  VbeStatus, AH
    POP  ES; POP  DI
    CMP  AX, 0x004F
    JE   ok
  }
  VbeErrorPtr(_AL);
  ok:
};

char VbeSetMode(int mode) {
  asm {
    PUSH BX; PUSH DI; PUSH ES
    MOV  AX, 0x4F02
    MOV  BX, mode
    LES  DI, crtinfoPtr
    INT  0x10
    MOV  VbeStatus, AH
    POP  ES; POP  DI; POP  BX
    CMP  AX, 0x004F
    JE   ok
  }
  VbeErrorPtr(_AL);
  ok:
};

char VbeGetMode(int *mode) {
  asm {
    PUSH BX
    MOV  AX, 0x4F03
    INT  0x10
    MOV  VbeStatus, AH
    MOV  mode, BX
    POP  BX
    CMP  AX, 0x004F
    JE   ok
  }
  VbeErrorPtr(_AL);
  ok:
};

char VbeSetDisplayWindow(char window, int winnumber) {
  asm {
    PUSH BX
    MOV  AX, 0x4F05
    MOV  BH, 0x00
    MOV  BL, window
    MOV  DX, winnumber
    INT  0x10
    MOV  VbeStatus, AH
    POP  BX
    CMP  AX, 0x004F
    JE   ok
  }
  VbeErrorPtr(_AL);
  ok:
};

char VbeGetDisplayWindow(char window, int *winnumber) {
  asm {
    PUSH BX
    MOV  AX, 0x4F05
    MOV  BH, 0x01
    MOV  BL, window
    INT  0x10
    MOV  VbeStatus, AH
    MOV  winnumber, DX
    POP  BX
    CMP  AX, 0x004F
    JE   ok
  }
  VbeErrorPtr(_AL);
  ok:
};
#pragma option -w+rvl

#endif