/*
      VGA
*/
#ifndef VGA_H
#define VGA_H

#pragma option -2

#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <alloc.h>
#include <memory.h>
#include <process.h>
#include <string.h>

#include <dtypes.h>

// use 32bit operations (only on 386+ processors)
#define USE_32BIT

#define VGA_WIDTH    320
#define VGA_HEIGHT   200
#define VGA_MAXX     VGA_WIDTH-1
#define VGA_MAXY     VGA_HEIGHT-1

// margin?

// drawMode constants
#define DM_NONE         0
#define DM_NORMAL       1
#define DM_ADD          2
#define DM_SUB1         3
#define DM_SUB2         4
#define DM_MUL          5
#define DM_AND          6
#define DM_OR           7
#define DM_XOR		8
#define DM_MIN          9
#define DM_MAX         10
#define DM_BLEND50     11
#define DM_BLEND       DM_BLEND50

BYTE far *videoseg = (BYTE far *) 0xA0000000L;

typedef short VGACOORD;
struct pointS {  // used by ftri()
  int x,y;
};

class VirtualScreen {
  public:
    char drawMode;
    BYTE far *bufferIn, far *bufferOut;
    VirtualScreen();
    VirtualScreen(VGACOORD, VGACOORD);
    VirtualScreen(VirtualScreen *);
    ~VirtualScreen();
    int error(int id);
    VGACOORD getWidth() { return(width); }
    VGACOORD getHeight() { return(height); }
    void restoreBuffers();
    void clear(BYTE color);
    void copyTo(VirtualScreen *virscr);
    void copyTo(VirtualScreen *virscr, VGACOORD, VGACOORD);
    void copyTo(VirtualScreen *virscr, BYTE skipCol);
    void copyRectTo(VirtualScreen *virscr, VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2);
    // ?? void copyRect(VirtualScreen *virscr, VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2);
    void getRect(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, BYTE *);
    void putRect(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, BYTE *);
    // far versions:
    void getRect(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, BYTE far *);
    void putRect(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, BYTE far *);
    void show();

    void putPixel(VGACOORD x, VGACOORD y, BYTE color);
    BYTE getPixel(VGACOORD x, VGACOORD y);
    void line(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, BYTE color);
    void rect(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, BYTE color);
    void circle(VGACOORD x, VGACOORD y, int rad, BYTE color);
    void fRect(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, BYTE color);
    void fCircle(VGACOORD x, VGACOORD y, int rad, BYTE color);
    void fTri(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, VGACOORD x3, VGACOORD y3, BYTE color);
    void text(VGACOORD x, VGACOORD y, char *data, BYTE size, BYTE color);

    void blur(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2);
    void pixelize(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, BYTE size);
    void stretch(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, VGACOORD x3, VGACOORD y3, VGACOORD x4, VGACOORD y4);
    void shiftUp(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, BYTE offset);
    void move(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, void transCoord(VGACOORD x,VGACOORD y, VGACOORD *xout,VGACOORD *yout));
  protected:
    VGACOORD width, height;
    DWORD bufferSize;
    BYTE far *screen;
    int checkCoords(VGACOORD *x1, VGACOORD *y1, VGACOORD *x2, VGACOORD *y2);

    int getLowY(pointS pts[3]);      //  \
    int getHighY(pointS pts[3]);     //   > used by ftri()
    void reorderX(pointS (*pts)[]);  //  /
};

VirtualScreen::VirtualScreen() {
  drawMode = DM_NORMAL;
  width = VGA_WIDTH;
  height = VGA_HEIGHT;
  bufferSize = (DWORD)width*height;
  screen = (BYTE far *)farmalloc(bufferSize);
  if (screen == NULL) error(0);
  bufferIn = screen;
  bufferOut = screen;
}

VirtualScreen::VirtualScreen(VGACOORD x, VGACOORD y) {
  drawMode = DM_NORMAL;
  width = x;
  height = y;
  bufferSize = (DWORD)width*height;
  screen = (BYTE far *)farmalloc(bufferSize);
  if (screen == NULL) error(0);
  bufferIn = screen;
  bufferOut = screen;
}

VirtualScreen::VirtualScreen(VirtualScreen *virscr) {
  drawMode = DM_NORMAL;
  width = virscr->getWidth();
  height = virscr->getHeight();
  bufferSize = (DWORD)width*height;
  screen = (BYTE far *)farmalloc(bufferSize);
  if (screen == NULL) error(0);
  bufferIn = screen;
  bufferOut = screen;
}

VirtualScreen::~VirtualScreen() {
  if (screen != NULL) farfree(screen);
}

int VirtualScreen::error(int id) {
  asm {
    mov  ax, 0x0003
    int  0x10
  }
  printf("ERROR %i",id);
  getch();
  abort();
  return(id);
}

void VirtualScreen::restoreBuffers() {
  bufferIn = screen;
  bufferOut = screen;
}

void VirtualScreen::clear(BYTE color) {
  #ifndef USE_32BIT
  _fmemset(screen,color,bufferSize);
  #else
  asm {
    PUSH ES
    PUSH DI
    mov  bx, this
    sub  bx, offset this
    DB 0x66; XOR  DI, DI    // xor  edi, edi
    DB 0x66; XOR  CX, CX    // xor  ecx, ecx
    LES  DI, [bx+offset this.bufferOut]
    MOV  AL, color
    MOV  AH, AL
    PUSH AX
    DB 0x66; SHL  AX, 16    // shl  eax, 16
    POP  AX
    db 0x66; MOV  CX, [bx+offset this.bufferSize]
    db 0x66; shr  cx, 2     // shr  ecx, 2
    DB 0x66; REP  STOSW     // rep stosd
    POP  DI
    POP  ES
  }
  #endif
}

void VirtualScreen::show() {
  #ifndef USE_32BIT
  _fmemcpy(videoseg, screen, bufferSize);
  #else
  asm {
    PUSH ES
    PUSH DS
    PUSH DI
    PUSH SI
    mov  bx, this
    sub  bx, offset this
    DB 0x66; XOR  DI, DI    // xor  edi, edi
    DB 0x66; XOR  SI, SI    // xor  esi, esi
    db 0x66; MOV  CX, [bx+offset this.bufferSize]
    db 0x66; shr  cx, 2     // shr  ecx, 2
    LES  DI, videoseg
    LDS  SI, [bx+offset this.bufferIn]
    DB 0x66; REP  MOVSW     // rep movsd
    POP  SI
    POP  DI
    POP  DS
    POP  ES
  }
  #endif
}

void VirtualScreen::copyTo(VirtualScreen *virscr) {
  #ifndef USE_32BIT
  _fmemcpy(virscr->screen, bufferIn, bufferSize);
  #else
  void far *scR = bufferIn, far *scW = virscr->screen;
  WORD siz = (bufferSize>>2);
  asm {
    PUSH ES
    PUSH DS
    PUSH DI
    PUSH SI
    DB 0x66; XOR  DI, DI    // xor  edi, edi
    DB 0x66; XOR  SI, SI    // xor  esi, esi
    DB 0x66; XOR  CX, CX    // xor  ecx, ecx
    LES  DI, scW
    LDS  SI, scR
    MOV  CX, siz
    DB 0x66; REP  MOVSW     // rep movsd
    POP  SI
    POP  DI
    POP  DS
    POP  ES
  }
  #endif
}

void VirtualScreen::copyTo(VirtualScreen *virscr, VGACOORD x, VGACOORD y) {
// copies actual screen to virscr, at pos x,y
  register int j;
  int dimx, dimy, bound,
      inx,iny, sizx,sizy;
  BYTE far *bufIn, far *bufOut;

  inx = (x<0)?-x:0;
  iny = (y<0)?-y:0;
  sizx = width-inx;
  sizy = height-iny;
  bound = x+inx+sizx-virscr->getWidth();
  dimx = sizx-((bound<0)?0:bound);
  bound = y+iny+sizy-virscr->getHeight();
  dimy = sizy-((bound<0)?0:bound);

  bufIn = bufferIn+iny*width+inx;
  bufOut = virscr->bufferOut+(y+iny)*virscr->getWidth()+x+inx;
  for (j=0; j<dimy; j++) {
    _fmemcpy(bufOut, bufIn, dimx);
    bufOut += virscr->getWidth();
    bufIn += width;
  }
}

void VirtualScreen::copyTo(VirtualScreen *virscr, BYTE skipCol) {
  void far *scR = bufferIn, far *scW = virscr->screen;
  WORD siz = bufferSize;
  asm {
    PUSH ES
    PUSH DS
    PUSH DI
    PUSH SI
    LES  DI, scW
    LDS  SI, scR
    MOV  CX, siz
    MOV	 AH, skipCol
  } skipLoop: asm {
    LODSB
    CMP	 AH, AL
    JNE  skip
    INC  DI
    JMP	 contLoop
  } skip: asm {
    STOSB
  } contLoop: asm {
    LOOP skipLoop
    POP  SI
    POP  DI
    POP  DS
    POP  ES
  }
}

void VirtualScreen::copyRectTo(VirtualScreen *virscr, VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2) {
  VGACOORD j, p;
  if (!checkCoords(&x1,&y1,&x2,&y2)) return;
  for (j=y1;j<=y2;j++)
    for (p=x1;p<=x2;p++) virscr->putPixel(p,j, *(bufferIn+j*width+p));
}

void VirtualScreen::getRect(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, BYTE *buffer) {
  VGACOORD j, c=0;
  if (!checkCoords(&x1,&y1,&x2,&y2)) return;
  for (j=y1;j<=y2;j++) {
    _fmemcpy(buffer+c, bufferIn+j*width+x1, x2-x1+1);
    c += x2-x1+1;
  }
}

void VirtualScreen::putRect(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, BYTE *buffer) {
  VGACOORD j, p, c, l;
  if (!checkCoords(&x1,&y1,&x2,&y2)) return;
  if (drawMode == DM_NORMAL) { // fast!!
    l = abs(x2-x1+1);
    for (j=y1,c=0,p=y1*width+x1; j<=y2; j++,p+=width,c+=l)
      _fmemcpy(screen+p, buffer+c, l);
  } else {
    for (j=y1;j<=y2;j++)
      for (p=x1;p<=x2;p++) putPixel(p,j, *(buffer+c++));
  }
}

// far versions
void VirtualScreen::getRect(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, BYTE far *buffer) {
  VGACOORD j, c=0;
  if (!checkCoords(&x1,&y1,&x2,&y2)) return;
  for (j=y1;j<=y2;j++) {
    _fmemcpy(buffer+c, bufferIn+j*width+x1, x2-x1+1);
    c += x2-x1+1;
  }
}

void VirtualScreen::putRect(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, BYTE far *buffer) {
  VGACOORD j, p, c, l;
  if (!checkCoords(&x1,&y1,&x2,&y2)) return;
  if (drawMode == DM_NORMAL) { // fast!!
    l = abs(x2-x1+1);
    for (j=y1,c=0,p=y1*width+x1; j<=y2; j++,p+=width,c+=l)
      _fmemcpy(screen+p, buffer+c, l);
  } else {
    for (j=y1;j<=y2;j++)
      for (p=x1;p<=x2;p++) putPixel(p,j, *(buffer+c++));
  }
}

void VirtualScreen::putPixel(VGACOORD x, VGACOORD y, BYTE color) {
  /*if ((x<0)||(y<0)||(x>=width)||(y>=height)) return;
  BYTE far *scrI = bufferIn+y*width+x,
       far *scrO = bufferOut+y*width+x;
  switch (drawMode) {
    case DM_ADD: *scrO = (*scrI+color>255?255:*scrI+color); break;
    case DM_SUB1: *scrO = (*scrI-color<0?0:*scrI-color); break;
    case DM_SUB2: *scrO = (color-*scrI<0?0:color-*scrI); break;
    case DM_MUL: *scrO = (*scrI*color>255?255:*scrI*color); break;
    case DM_OR: *scrO = *scrI&color; break;
    case DM_AND: *scrO = *scrI|color; break;
    case DM_XOR: *scrO = *scrI^color; break;
    case DM_MIN: *scrO = min(*scrI,color); break;
    case DM_MAX: *scrO = max(*scrI,color); break;
    case DM_BLEND50: *scrO = (*scrI+color)>>1; break;
    case DM_NONE: break;
    case DM_NORMAL: ;
    default: *scrO = color;
  }*/
  asm {
    push es
    push ds
    //push di
    //push si
    push bx
    push cx

    mov  si, this
    sub  si, offset this
    mov  bx, x          // load coords
    mov  cx, y

    cmp  bx, 0          // check coords
    jnl  putPixel_inv01
    jmp  putPixel_end
  } putPixel_inv01: asm {
    cmp  cx, 0
    jnl  putPixel_inv02
    jmp  putPixel_end
  } putPixel_inv02: asm {
    cmp  bx, [si+offset this.width]
    jl   putPixel_inv03
    jmp  putPixel_end
  } putPixel_inv03: asm {
    cmp  cx, [si+offset this.height]
    jl   putPixel_inv04
    jmp  putPixel_end
  } putPixel_inv04: asm {

    mov  ax, cx         // calc buffer offset
    mul  word ptr [si+offset this.width]
    add  ax, bx
    mov  bl, [si+offset this.drawMode]
    les  di, [si+offset this.bufferOut]
    lds  si, [si+offset this.bufferIn]
    add  di, ax
    add  si, ax

    mov  ah, bl
    mov  al, color
    cmp  ah, DM_BLEND50
    ja   putPixel_end
    jne  putPixel_inv1      // inverse jump (>128)
    jmp  putPixel_dm_blend50
  } putPixel_inv1: asm {
    cmp  ah, DM_AND  // 6
    jne  putPixel_inv2
    jmp  putPixel_dm_and
  } putPixel_inv2: asm {
    jb   putPixel_dmless6
    ja   putPixel_dmmore6
  } putPixel_dmless6: asm {
    cmp  ah, DM_SUB1  // 3
    jne  putPixel_inv3
    jmp  putPixel_dm_sub1
  } putPixel_inv3: asm {
    jb   putPixel_dmless3
    ja   putPixel_dmmore3
  } putPixel_dmless3: asm {
    cmp  ah, DM_NORMAL  // 1
    jne  putPixel_inv4
    jmp   putPixel_dm_norm
  } putPixel_inv4: asm {
    jb   putPixel_end
    ja   putPixel_dm_add
  } putPixel_dmmore3: asm {
    cmp  ah, DM_SUB2  // 4
    jne  putPixel_inv5
    jmp   putPixel_dm_sub2
  } putPixel_inv5: asm {
    ja   putPixel_dm_mul
  } putPixel_dmmore6: asm {
    cmp  ah, DM_MIN  // 9
    jne  putPixel_inv6
    jmp   putPixel_dm_min
  } putPixel_inv6: asm {
    jb   putPixel_dmless9
    ja   putPixel_dm_max
  } putPixel_dmless9: asm {
    cmp  ah, DM_OR  // 7
    jne  putPixel_inv7
    jmp   putPixel_dm_or
  } putPixel_inv7: asm {
    ja   putPixel_dm_xor
  } putPixel_dm_add: asm {      // DM_ADD
    add  al, ds:[si]
    jnc  putPixel_dm_add_nc
    mov  al, 0xFF
  } putPixel_dm_add_nc: asm {
    mov  es:[di], al
    jmp  putPixel_end
  } putPixel_dm_sub1: asm {     // DM_SUB1
    mov  dl, ds:[si]
    sub  dl, al
    jnb  putPixel_dm_sub1_nb
    mov  dl, 0x00
  } putPixel_dm_sub1_nb: asm {
    mov  es:[di], dl
    jmp  putPixel_end
  } putPixel_dm_sub2: asm {     // DM_SUB2
    sub  al, ds:[si]
    jnb  putPixel_dm_sub2_nb
    mov  al, 0x00
  } putPixel_dm_sub2_nb: asm {
    mov  es:[di], al
    jmp  putPixel_end
  } putPixel_dm_mul: asm {      // DM_MUL
    mov  dl, ds:[si]
    mul  dl
    cmp  ah, 0
    je   putPixel_dm_mul_nc
    mov  al, 0xFF
  } putPixel_dm_mul_nc: asm {
    mov  es:[di], al
    jmp  putPixel_end
  } putPixel_dm_and: asm {      // DM_AND
    and  al, ds:[si]
    mov  es:[di], al
    jmp  putPixel_end
  } putPixel_dm_or: asm {       // DM_OR
    or   al, ds:[si]
    mov  es:[di], al
    jmp  putPixel_end
  } putPixel_dm_xor: asm {      // DM_XOR
    xor  al, ds:[si]
    mov  es:[di], al
    jmp  putPixel_end
  } putPixel_dm_min: asm {      // DM_MIN
    mov  dl, ds:[si]
    cmp  al, dl
    jbe  putPixel_dm_min_b
    mov  al, dl
  } putPixel_dm_min_b: asm {
    mov  es:[di], al
    jmp  putPixel_end
  } putPixel_dm_max: asm {      // DM_MAX
    mov  dl, ds:[si]
    cmp  al, dl
    jae  putPixel_dm_max_c
    mov  al, dl
  } putPixel_dm_max_c: asm {
    mov  es:[di], al
    jmp  putPixel_end
  } putPixel_dm_blend50: asm {  // DM_BLEND50
    mov  dl, ds:[si]
    xor  ah, ah
    xor  dh, dh
    add  ax, dx
    shr  ax, 1
  } putPixel_dm_norm: asm {     // DM_NORMAL
    mov  es:[di], al
  } putPixel_end: asm {
    pop  cx
    pop  bx
    //pop  si
    //pop  di
    pop  ds
    pop  es
  }
}

BYTE VirtualScreen::getPixel(VGACOORD x, VGACOORD y) {
  if ((x<0)||(y<0)||(x>=width)||(y>=height)) return(0);
  return( *(bufferIn+y*width+x) );
}

void VirtualScreen::line(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, BYTE color) {
  int dx = x2-x1, dy = y2-y1;
  /*if (dy == 0) { // horizontal line
    if (drawMode == DM_NORMAL) { // direct copy
      void far *sc = bufferOut;
      int maxx = width-1,
          maxy = height-1,
          wid = width;
      asm {
        push es
        push di
        push bx
        push cx
        push dx
        mov  ax, x1
        mov  bx, x2
        cmp  ax, bx
        jle  okOrder
        xchg ax, bx
      } okOrder: asm {
        cmp  ax, 0
        jge  okZero
        xor  ax, ax
      } okZero: asm {
        cmp  ax, maxx
        jg   endfunc
        cmp  bx, maxx
        jle  okLen
        mov  bx, maxx
      } okLen: asm {
        cmp  bx, 0
        jl   endfunc
        mov  cx, bx
        sub  cx, ax
        inc  cx
        les  di, sc
        mov  bx, ax
        mov  ax, y1
        cmp  ax, maxy
        jg   endfunc
        mov  dx, wid
        mul  dx
        add  ax, bx
        add  di, ax
        mov  al, color
        mov  ah, al
        test cx, 0x01
        jz   isEven
        stosb
       } isEven: asm {
        shr  cx, 1
        rep  stosw
       } endfunc: asm {
        pop  dx
        pop  cx
        pop  bx
        pop  di
        pop  es
      }
      return;
    }
  }
  */
  int ix = abs(dx), iy = abs(dy),
      inct = (ix>iy?ix:iy),
      plotX = x1, plotY = y1,
      x = 0, y = 0,
      i;
  char plot;

  putPixel(plotX,plotY, color);
  for (i=0;i<=inct;i++) {
    x += ix;
    y += iy;
    plot = 0;
    if (x > inct) {
      plot = 1;
      x -= inct;
      if (dx > 0) plotX++;
      if (dx < 0) plotX--;
     }
     if (y > inct) {
       plot = 1;
       y -= inct;
       if (dy > 0) plotY++;
       if (dy < 0) plotY--;
     }
     if (plot) putPixel(plotX,plotY, color);
  }
}

void VirtualScreen::rect(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, BYTE color) {
  if (!checkCoords(&x1,&y1,&x2,&y2)) return;
  line(x1,y1,x2,y1, color);
  line(x1,y1,x1,y2, color);
  line(x2,y1,x2,y2, color);
  line(x1,y2,x2,y2, color);
}

void VirtualScreen::circle(VGACOORD x, VGACOORD y, int rad, BYTE color) {
  register int d, nx,ny;
  int x1,x2,x3,x4, y1,y2,y3,y4;
  d = 3 - (rad<<1);
  nx = 0;
  ny = rad;
  x1 = x+nx;  x2 = x-nx;  x3 = x+ny;  x4 = x-ny;
  y1 = y+ny;  y2 = y-ny;  y3 = y+nx;  y4 = y-nx;
  while (nx <= ny) {
    putPixel(x1, y1, color);
    putPixel(x1, y2, color);
    if (x2 != x1) {
      putPixel(x2, y1, color);
      putPixel(x2, y2, color);
    }
    if (y1 != y3) {
      putPixel(x3, y3, color);
      putPixel(x4, y3, color);
      if (y3 != y4) {
        putPixel(x3, y4, color);
        putPixel(x4, y4, color);
      }
    }
    nx++;
    x1++;  x2--;
    y3++;  y4--;
    if (d < 0) {
      d += (nx<<2)+6;
    } else {
      d += ((nx-ny--)<<2)+10;
      x3--;  x4++;
      y1--;  y2++;
    }
  }
}

void VirtualScreen::fRect(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, BYTE color) {
  VGACOORD j, p;
  if (!checkCoords(&x1,&y1,&x2,&y2)) return;
  if (drawMode == DM_NORMAL) {
    for (j=y1,p=y1*width+x1; j<=y2; j++,p+=width)  // fast!!
      _fmemset(screen+p, color, abs(x2-x1+1));
  } else {
    for (j=y1;j<=y2;j++)
      for (p=x1;p<=x2;p++) putPixel(p,j,color);
  }
}

void VirtualScreen::fCircle(VGACOORD x, VGACOORD y, int rad, BYTE color) {
  register int d, nx,ny;
  int x1,x2,x3,x4, y1,y2,y3,y4;
  d = 3 - (rad<<1);
  nx = 0;
  ny = rad;
  x1 = x+nx;  x2 = x-nx;  x3 = x+ny;  x4 = x-ny;
  y1 = y+ny;  y2 = y-ny;  y3 = y+nx;  y4 = y-nx;
  int oldy1 = y1+1;
  while (nx <= ny) {
    if (y1 != oldy1) {
      line(x2,y1, x1,y1, color);
      line(x2,y2, x1,y2, color);
    }
    if (y3 != y1){
      line(x3,y3, x4,y3, color);
      if (y3 != y4) line(x3,y4, x4,y4, color);
    }
    oldy1 = y1;
    nx++;
    x1++;  x2--;
    y3++;  y4--;
    if (d < 0) {
      d += (nx<<2)+6;
    } else {
      d += ((nx-ny--)<<2)+10;
      x3--;  x4++;
      y1--;  y2++;
    }
  }
}

void VirtualScreen::fTri(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, VGACOORD x3, VGACOORD y3, BYTE color) {
// 287 processor needed!
  int a,b,c, t,k = 0;
  int x,y;
  float q[5],m[5];
  pointS pts[5], tl,br;

  pts[0].x = x1; pts[0].y = y1;
  pts[1].x = x2; pts[1].y = y2;
  pts[2].x = x3; pts[2].y = y3;
  reorderX(&pts);
  pts[3] = pts[0];
  pts[4] = pts[1];
  tl.x = pts[0].x; tl.y = pts[getLowY(pts)].y;
  br.x = pts[2].x; br.y = pts[getHighY(pts)].y;

  for (x=0;x<3;x++) {
    if (pts[x+1].x == pts[x].x) { k = 2+x; continue; }
    asm {
      push  di
      push  si
      lea   si, pts
      lea   di, m
      mov   bx, x
      shl   bx, 2
    }
    __emit__(0x9B,0xDB,0xE3); // finit
    asm {
      //m[x] = (float)(pts[x+1].y-pts[x].y)/(pts[x+1].x-pts[x].x);
      fild  word ptr [si+bx+4+0]  // x
      fisub word ptr [si+bx+0]
      fild  word ptr [si+bx+4+2]  // y
      fisub word ptr [si+bx+2]
      fdiv  st, st(1)
      fst   dword ptr [di+bx]
      //q[x] = pts[x].y-m[x]*pts[x].x;
      fild  word ptr [si+bx]
      fmul  dword ptr [di+bx]
      fchs
      fiadd word ptr [si+bx+2]
      lea   si, q
      fst   dword ptr [si+bx]
      pop   si
      pop   di
    }
    //if ((k == 0)&&(m[x] == 0)) k = 1;
  }
  asm {
    //m[3] = m[0];  m[4] = m[1];
    //q[3] = q[0];  q[4] = q[1];
    push  si
    push  di
    push  ds
    pop   es
    lea   di, m
    mov   si, di
    add   di, 4*3
    mov   cx, (4*2) shr 1
    rep   movsw
    lea   di, q
    mov   si, di
    add   di, 4*3
    mov   cx, (4*2) shr 1
    rep   movsw
    pop   di
    pop   si
  }

  if (k >= 2) { // vertical!
    k -= 2;
    t = 2;
  } else {
    k = 1;
    __emit__(0x9B,0xDB,0xE3); // finit
    asm {
      push  di
      lea   di, m
      fld   dword ptr [di]
      fcom  dword ptr [di+1*4]
    }
    __emit__(0xDF,0xE0); // fstsw ax
    asm {
      pop   di
      mov   t, 0x0000
      and   ax, 0100010100000000b
      cmp   ax, 0000000100000000b
      je    t1
      cmp   ax, 0100000000000000b
      je    t1
      jmp   t2
    }
    t1: t = 1;
    t2:
  }

  for (y=tl.y;y<=br.y;y++) {
    for (x=tl.x;x<=br.x;x++) {
      asm {
	push  si
	push  di
	mov   bx, k
	shl   bx, 2
	lea   si, q
	lea   di, m
      }
      __emit__(0x9B,0xDB,0xE3); // finit
      asm {
	fild  x
	fmul  dword ptr [di+bx+2*4]
	fadd  dword ptr [si+bx+2*4]
	fisub y
	fist  a
	fild  x
	fmul  dword ptr [di+bx]
	fadd  dword ptr [si+bx]
	fisub y
	fist  b
	fild  x
	fmul  dword ptr [di+bx+1*4]
	fadd  dword ptr [si+bx+1*4]
	fisub y
	fist  c
	pop  di
	pop  si
      }
      switch (t) {
	case 0: {
	  a = (a >= 0);  b = (b >= 0);  c = (c <= 0);
	  break;
	}
	case 1: {
	  a = (a <= 0);  b = (b <= 0);  c = (c >= 0);
	  break;
	}
	case 2: {
	  a = (a <= 0);  b = 1;  c = (c >= 0);
	  break;
	}
      }
      if (a && b && c) putPixel(x,y, color);
    }
  }
}

void VirtualScreen::text(VGACOORD x, VGACOORD y, char *data, BYTE size, BYTE color) {
  int i,j,k;
  BYTE chr;
  static BYTE far *charTab = 0;
  if (charTab == 0) {
    asm {
      push  es
      push  bp
      mov   ax, 0x1130
      mov   bh, 0x06
      int   0x10
      mov   word [charTab], bp   // only if static
      mov   word [charTab+2], es
      pop   bp
      pop   es
    }
  }
  for (i=0;i<strlen(data);i++) {
    for (j=0;j<16;j++) {
      chr = *(charTab+(BYTE)data[i]*16+j);
      for (k=0;k<8;k++) {
	asm ROL chr, 1;
	if (chr & 0x01)
	  if (size>1) fRect(x+k*size+i*8*size,y+j*size,x+k*size+i*8*size+size-1,y+j*size+size-1,color); else
	    putPixel(x+k+i*8,y+j,color);
      }
    }
  }
}

void VirtualScreen::blur(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2) {
  BYTE b1,b2,b3,b4;
  int c,m;
  BYTE far *scrI;
  register int i, j;
  if (!checkCoords(&x1,&y1,&x2,&y2)) return;
  j = y1;
  scrI = bufferIn+j*width+x1;
  m = j*width;
  for (i=x1;i<=x2;i++) {              // 1st line
    b2 = ((i==x1) ? 0 : *(scrI-1));
    b3 = ((i>=x2) ? 0 : *(scrI+1));
    b4 = *(scrI+width);
    c = b2+b3+b4;
    *(bufferOut+m+i) = c >> 2;
    scrI++;
  }
  for (j=y1;j<y2;j++) {
    scrI = bufferIn+j*width+x1;
    m = (j+1)*width;
    for (i=x1;i<=x2;i++) {
      b1 = *scrI;
      b2 = ((i==x1) ? 0 : *(scrI+width-1));
      b3 = ((i>=x2) ? 0 : *(scrI+width+1));
      b4 = ((j>=y2-1) ? 0 : *(scrI+(width<<1)));
      c = b1+b2+b3+b4;
      *(bufferOut+m+i) = c >> 2;
      scrI++;
    }
  }
}

void VirtualScreen::pixelize(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, BYTE size) {
  int i,j;
  BYTE tcol;
  if (!checkCoords(&x1,&y1,&x2,&y2)||size<2) return;
  for (j=y1;j<=y2;j += size)
    for (i=x1;i<=x2;i += size) {
      tcol = *(bufferIn+j*width+i);
      fRect(i,j,i+size-1,j+size-1,tcol);
    }
}

void VirtualScreen::stretch(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, VGACOORD x3, VGACOORD y3, VGACOORD x4, VGACOORD y4) {
  BYTE tcol;
  unsigned int i,j, k,l;
  if (!checkCoords(&x1,&y1,&x2,&y2)) return;
  l = y1;
  for (j=y3;j<y4;j++) {
    k = x1;
    for (i=x3;i<x4;i++) {
      tcol = *(bufferIn+l*width+k);
      putPixel(i,j,tcol);
      k = (long)(i-x3)*(x2-x1)/(x4-x3)+x1;
    }
    l = (j-y3)*(y2-y1)/(y4-y3)+y1;
  }
}

void VirtualScreen::shiftUp(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, BYTE offset) {
  int i,j;
  if (!checkCoords(&x1,&y1,&x2,&y2)) return;
  for (j=y1;j<=y2-offset;j++) {
    _fmemcpy(bufferOut+j*width+x1, bufferIn+(j+offset)*width+x1, x2-x1);
  }
}

#pragma option -w-ccc
void VirtualScreen::move(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2, void transCoord(VGACOORD x,VGACOORD y, VGACOORD *xout,VGACOORD *yout)) {
  VGACOORD i,j, xn,yn;
  if (!checkCoords(&x1,&y1,&x2,&y2)) return;
  for (j=y1;j<=y2;j++) {
    for (i=x1;i<=x2;i++) {
      BYTE color;
      transCoord(i,j, &xn,&yn);
      if ((xn<0)||(yn<0)||(xn>=width)||(yn>=height))
	//{xn %= VGA_WIDTH; yn %= VGA_HEIGHT; }
	{ color = 0;}
      else
	color = *(bufferIn+yn*width+xn);
      putPixel(i,j, color);
    }
  }
}
#pragma option -w+ccc

int VirtualScreen::getLowY(pointS pts[3]) {
  int i, a = 0;
  for (i=1;i<3;i++) if (pts[a].y > pts[i].y) a = i;
  return a;
}

int VirtualScreen::getHighY(pointS pts[3]) {
  int i, a = 0;
  for (i=1;i<3;i++) if (pts[a].y < pts[i].y) a = i;
  return a;
}

void xchg(pointS *a, pointS *b) {
  asm {
    push  si
    push  di
    mov   si, a
    mov   di, b
    mov   ax, [si]
    mov   dx, [si+2]
    xchg  [di],ax
    xchg  [di+2],dx
    mov   [si], ax
    mov   [si+2], dx
    pop   di
    pop   si
  }
}

void VirtualScreen::reorderX(pointS (*pts)[]) {
  int i, j;
  for (j=0;j<2;j++) {
    for (i=0;i<2;i++) {
      if ((*pts)[i].x >= (*pts)[i+1].x) {
	if ((*pts)[i].x == (*pts)[i+1].x) {
	  if ((*pts)[i].y > (*pts)[i+1].y) xchg(&(*pts)[i+1], &(*pts)[i]);
	} else xchg(&(*pts)[i+1], &(*pts)[i]);
      }
    }
  }
}

int VirtualScreen::checkCoords(VGACOORD *x1, VGACOORD *y1, VGACOORD *x2, VGACOORD *y2) {
  VGACOORD t;
  order(x1,x2);
  order(y1,y2);
  if (*x1<0) *x1 = 0;
  if (*y1<0) *y1 = 0;
  if (*x2>=width) *x2 = width-1;
  if (*y2>=height) *y2 = height-1;
  if ((*x1>=width)||(*y1>=height)||(*x2<0)||(*y2<0)) return(0);
  return(1);
}

//---------------------------------------------------------------------------
#endif
