/* VGAPAL library for C++ */

#ifndef VGAPAL_H
#define VGAPAL_H

#include <stdlib.h>
#include <dtypes.h>
#include <io.h>
#include <fcntl.h>

typedef struct {
  BYTE align, b,g,r;
} RGBcol;
typedef RGBcol pal256[256];

const RGBcol COL_BLACK = {0x00,0x00,0x00,0},
	     COL_WHITE = {0x1F,0x3F,0x1F,0},
	     COL_RED =   {0x00,0x00,0x1F,0};

void pal_set(pal256 *palette, BYTE ncol, BYTE r, BYTE g, BYTE b) {
  (*palette)[ncol].r = r;
  (*palette)[ncol].g = g;
  (*palette)[ncol].b = b;
}

void pal_in(pal256 *palette, BYTE ncol) {
  BYTE r,g,b;
  asm {
    PUSH DX
    MOV  DX, 0x03C7
    MOV  AL, ncol
    OUT  DX, AL
    ADD  DX, 0x02
    IN   AL, DX
    MOV  r , AL
    IN   AL, DX
    MOV  g , AL
    IN   AL, DX
    MOV  b , AL
    POP  DX
  }
  (*palette)[ncol].r = r;
  (*palette)[ncol].g = g;
  (*palette)[ncol].b = b;
}

void pal_out(pal256 *palette, BYTE ncol) {
  BYTE r,g,b;
  r = (*palette)[ncol].r;
  g = (*palette)[ncol].g;
  b = (*palette)[ncol].b;
  asm {
    PUSH DX
    MOV  DX, 0x03C8
    MOV  AL, ncol
    OUT  DX, AL
    INC  DX
    MOV  AL, r
    OUT  DX, AL
    MOV  AL, g
    OUT  DX, AL
    MOV  AL, b
    OUT  DX, AL
    POP  DX
  }
}

void pal_inRGB(BYTE ncol, BYTE *r, BYTE *g, BYTE *b) {
  BYTE rl,gl,bl;
  asm {
    PUSH DX
    MOV  DX, 0x03C7
    MOV  AL, ncol
    OUT  DX, AL
    ADD  DX, 0x02
    IN   AL, DX
    MOV  rl , AL
    IN   AL, DX
    MOV  gl , AL
    IN   AL, DX
    MOV  bl , AL
    POP  DX
  }
  *r = rl;
  *g = gl;
  *b = bl;
}


void pal_outRGB(BYTE ncol, BYTE r, BYTE g, BYTE b) {
  asm {
    PUSH DX
    MOV  DX, 0x03C8
    MOV  AL, ncol
    OUT  DX, AL
    INC  DX
    MOV  AL, r
    OUT  DX, AL
    MOV  AL, g
    OUT  DX, AL
    MOV  AL, b
    OUT  DX, AL
    POP  DX
  }
}

void pal_inAll(pal256 *palette) {
  for (int ncol=0;ncol<=255;ncol++)
    pal_in(palette,ncol);
}

void pal_outAll(pal256 *palette) {
  for (int ncol=0;ncol<=255;ncol++)
    pal_out(palette,ncol);
}

void pal_setGrey(pal256 *palette, BYTE start, BYTE end) {
  int i,c;
  for (i=start;i<=end;i++) {
    c = ((*palette)[i].r+(*palette)[i].g+(*palette)[i].b) / 3;
    (*palette)[i].r = (*palette)[i].g = (*palette)[i].b = c;
  }
}

void pal_setGradient(pal256 *palette, BYTE start,BYTE end, BYTE r1,BYTE g1,BYTE b1, BYTE r2,BYTE g2,BYTE b2) {
  WORD a = 0;
  BYTE b = start, n = end-start;
  if (start >= end) return;
  do {
    a++;
    (*palette)[b].r = r1 - r1*a/n + r2*a/n;
    (*palette)[b].g = g1 - g1*a/n + g2*a/n;
    (*palette)[b].b = b1 - b1*a/n + b2*a/n;
    b++;
  } while ((a<end+1)&&(b<end));
  (*palette)[b].r = r2;
  (*palette)[b].g = g2;
  (*palette)[b].b = b2;
}

void pal_invert(pal256 *palette, BYTE start, BYTE end) {
  int ncol;
  for (ncol=start;ncol<=end;ncol++) {
    (*palette)[ncol].r = 255-(*palette)[ncol].r;
    (*palette)[ncol].g = 255-(*palette)[ncol].g;
    (*palette)[ncol].b = 255-(*palette)[ncol].b;
  }
}

void pal_light(pal256 *palette, BYTE start, BYTE end, BYTE step) {
  int i;
  for (i=start;i<=end;i++) {
    (*palette)[i].r = min((short)(*palette)[i].r+step, 63);
    (*palette)[i].g = min((short)(*palette)[i].g+step, 63);
    (*palette)[i].b = min((short)(*palette)[i].b+step, 63);
  }
}

void pal_dark(pal256 *palette, BYTE start, BYTE end, BYTE step) {
  int i;
  for (i=start;i<=end;i++) {
    (*palette)[i].r = max((int)(*palette)[i].r-step, 0);
    (*palette)[i].g = max((int)(*palette)[i].g-step, 0);
    (*palette)[i].b = max((int)(*palette)[i].b-step, 0);
  }
}

/*
Procedure ShlPal(Var Palette: Pal256; StartVal,EndVal, Step: Byte);
Var a: Byte;
  Begin
    For a := StartVal To EndVal Do
      Begin
	If a = EndVal Then GetPal(Palette, StartVal) Else GetPal(Palette, a+Step);
	Palette[a].R := R;
	Palette[a].G := G;
	Palette[a].B := B;
      End;
  End;

Procedure ShrPal(Var Palette: Pal256; StartVal,EndVal, Step: Byte);
Var a: Byte;
  Begin
    For a := EndVal DownTo StartVal Do
      Begin
	If a = StartVal Then GetPal(Palette, EndVal) Else GetPal(Palette, a-Step);
	Palette[a].R := R;
        Palette[a].G := G;
        Palette[a].B := B;
      End;
  End;


Procedure DecR(Var Palette: Pal256; StartC,EndC, Step: Byte);
Var a: Byte;
  Begin
    For a := StartC To EndC Do
      Begin
        If Palette[a].R < Step Then Palette[a].R := 0 Else Palette[a].R := Palette[a].R-Step;
      End;
  End;

Procedure DecG(Var Palette: Pal256; StartC,EndC, Step: Byte);
Var a: Byte;
  Begin
    For a := StartC To EndC Do
      Begin
	If Palette[a].G < Step Then Palette[a].G := 0 Else Palette[a].G := Palette[a].G-Step;
      End;
  End;

Procedure DecB(Var Palette: Pal256; StartC,EndC, Step: Byte);
Var a: Byte;
  Begin
    For a := StartC To EndC Do
      Begin
        If Palette[a].B < Step Then Palette[a].B := 0 Else Palette[a].B := Palette[a].B-Step;
      End;
  End;

Procedure IncR(Var Palette: Pal256; StartC,EndC, Step: Byte);
Var a: Byte;
  Begin
    For a := StartC To EndC Do
      Begin
        If Palette[a].R > 63-Step Then Palette[a].R := 255 Else Palette[a].R := Palette[a].R+Step;
      End;
  End;

Procedure IncG(Var Palette: Pal256; StartC,EndC, Step: Byte);
Var a: Byte;
  Begin
    For a := StartC To EndC Do
      Begin
        If Palette[a].G > 63-Step Then Palette[a].G := 255 Else Palette[a].G := Palette[a].G+Step;
      End;
  End;

Procedure IncB(Var Palette: Pal256; StartC,EndC, Step: Byte);
Var a: Byte;
  Begin
    For a := StartC To EndC Do
      Begin
	If Palette[a].B > 63-Step Then Palette[a].B := 255 Else Palette[a].B := Palette[a].B+Step;
      End;
  End;


Procedure CopyPal(Var Palette1,Palette2: Pal256; StartC,EndC: Byte);
Var a: Byte;
  Begin
    For a := StartC To EndC Do
      Begin
	Palette2[a].R := Palette1[a].R;
	Palette2[a].G := Palette1[a].G;
	Palette2[a].B := Palette1[a].B;
      End;
  End;
*/

void pal_setRandom(pal256 *palette, BYTE maxcol) {
  BYTE col, n;
  BYTE rgb1[3], rgb2[3];
  col = random(maxcol)+1;
  rgb2[0] = random(64);
  rgb2[1] = random(64);
  rgb2[2] = random(64);
  for (n=1;n<=col;n++) {
    rgb1[0] = random(64);
    rgb1[1] = random(64);
    rgb1[2] = random(64);
    pal_setGradient(palette, 256/col*(n-1),256/col*n-1,
      rgb2[0],rgb2[1],rgb2[2], rgb1[0],rgb1[1],rgb1[2]);
    if (n==col) break;
    rgb2[0] = rgb1[0];
    rgb2[1] = rgb1[1];
    rgb2[2] = rgb1[2];
  }
  pal_setGradient(palette, 256/col*(n-1),255, rgb2[0],rgb2[1],rgb2[2],rgb1[0],rgb1[1],rgb1[2]);
}

typedef struct {
  DWORD signature;
  DWORD fileSize;
  DWORD type;
} RIFFHead;

typedef struct {
  DWORD signature;
  DWORD chunkSize;
} RIFFChunk;

int pal_save(pal256 *palette, BYTE start, BYTE end, char *fname) {
// CORRECT ME!
  int colors = end-start+1;
  int fsize = sizeof(RIFFHead)+sizeof(RIFFChunk)+4*(colors+1);
  DWORD data;
  RIFFHead rh = {0x52494646, fsize-8, 0x50414C20};
  RIFFChunk rc = {0x64617461, fsize-20};
  data = (colors << 16) | 0x0300;

  int handle = _open(fname, O_WRONLY);
  if (handle == -1) return(0);
  _write(handle, &rh, sizeof(rh));
  _write(handle, &rc, sizeof(rc));
  _write(handle, &data, sizeof(data));
  for (int i=start; i<=end; i++) {
    _write(handle, (DWORD *)(palette[i]), 4);
  }
  _close(handle);
  return(1);
}

int pal_load(pal256 *palette, BYTE start, BYTE end, char *fname) {
// CORRECT ME!
  int colors;
  DWORD data;
  RIFFHead rh;
  RIFFChunk rc;

  int handle = _open(fname, O_RDONLY);
  if (handle == -1) return(0);
  _read(handle, &rh, sizeof(rh));
  _read(handle, &rc, sizeof(rc));
  _read(handle, &data, sizeof(data));
  colors = min(data >> 16, end-start+1);
  for (int i=0; i<colors; i++) {
    _read(handle, (DWORD *)(palette[i+start]), 4);
  }
  _close(handle);
  return(1);
}
#endif
