#ifndef W3D_DECL
  #include <w3d\w3d_decl.h>
#endif
#ifdef W3D_OBJECT
  #error W3d object already included.
#endif
#define W3D_OBJECT
#include <alloc.h>
#include <string.h>

class object3D:public obj_p {           // classe che gestisce gli object3D
  private:
  point *pointH;
  point3Ds *lastNode;
  edge *lastEdge;
  point3Ds *getNextNode();
  edge *object3D::getNextEdge();

  public:
  int drawMode;     // 0 = points, 1 = lines, 2 = planes (nd)
  object3D(point *ph);
  void init(object3Ds *obj);
  void setName(object3Ds *obj, char *objName);
  char *getName(object3Ds *obj);
  point3Ds *getNode(object3Ds *obj, int index);
  void addNode(object3Ds *obj, point3Ds p3d);
  void delNode(object3Ds *obj, int index);
  edge *getEdge(object3Ds *obj, int index);
  void addEdge(object3Ds *obj, edge edg);
  void delEdge(object3Ds *obj, int index);
  void move(object3Ds *obj, point3Ds p3d);
  void scaleO(object3Ds *obj, double scalX, double scalY, double scalZ);
  void rotateO(object3Ds *obj, double alpha, double beta, double gamma);
  void rotateP(object3Ds *obj, point3Ds center, double alpha, double beta, double gamma);
  void draw(object3Ds *obj);
  object3Ds *duplicate(object3Ds *obj);
  //void clear(object3Ds *obj);
};

object3D::object3D(point *ph) {
  pointH = ph;
  parentWld = NULL;
  lastNode = NULL;
  lastEdge = NULL;
  drawMode = 1;
}

void object3D::init(object3Ds *obj) {
  obj->name = (char*)malloc(1);
  obj->name = 0;
  obj->cNodes = 0;
  obj->cEdges = 0;
  obj->nodes = NULL;
  obj->edges = NULL;
}

void object3D::setName(object3Ds *obj, char *objName) {
  obj->name = (char *)realloc((void *)obj->name,strlen(objName)+1);
  //if (obj->name == NULL) { printf("memory error\n"); }
  obj->name = objName;
}

char *object3D::getName(object3Ds *obj) {
  return(obj->name);
}

point3Ds *object3D::getNode(object3Ds *obj, int index) {
  int i;
  point3Ds *lastpnt;
  if (index >= obj->cNodes) return(NULL);
  lastpnt = obj->nodes;
  for (i=0;i<index;i++) {
    asm {
      PUSH SI
      MOV  SI, lastpnt
      ADD  SI, SIZE_P3D
      MOV  AX, DS:[SI][0x04]
      MOV  lastpnt, AX
      POP  SI
    }
  }
  lastNode = lastpnt;
  return(lastpnt);
}

point3Ds *object3D::getNextNode() {
  point3Ds *next;
  next = lastNode;
  if (next == NULL) return(NULL);
  asm {
    PUSH SI
    MOV  SI, next
    ADD  SI, SIZE_P3D
    MOV  AX, DS:[SI][0x04]
    MOV  next, AX
    POP  SI
  }
  lastNode = next;
  return(next);
}

void object3D::addNode(object3Ds *obj, point3Ds p3d) {
  point3Ds *prevN, *nextN;
  point3Ds *newpnt = (point3Ds *)malloc(sizeof(point3Ds)+8);
  if (!obj->cNodes) {
    prevN = NULL;
    obj->nodes = newpnt;
  } else {
    prevN = getNode(obj,obj->cNodes-1);
    asm {
      PUSH SI
      MOV  SI, prevN
      ADD  SI, SIZE_P3D
      MOV  AX, newpnt
      MOV  DS:[SI][0x04], AX
      POP  SI
    }
  }
  nextN = NULL;
  asm {
    PUSH SI
    MOV  SI, newpnt
    ADD  SI, SIZE_P3D
    MOV  AX, prevN
    MOV  DS:[SI], AX
    MOV  AX, nextN
    MOV  DS:[SI][0x04], AX
    POP  SI
  }
  newpnt->x = p3d.x;
  newpnt->y = p3d.y;
  newpnt->z = p3d.z;
  obj->cNodes++;
}

void object3D::delNode(object3Ds *obj, int index) {
  if (index >= obj->cNodes) return;
  point3Ds *prevN, *thisN, *nextN;
  thisN = getNode(obj, index);
  asm {
    PUSH SI
    MOV  SI, thisN
    ADD  SI, SIZE_P3D
    MOV  AX, DS:[SI]
    MOV  prevN, AX
    MOV  AX, DS:[SI][0x04]
    MOV  nextN, AX
    POP  SI
  }
  if (prevN)
    asm {
      PUSH SI
      MOV  SI, prevN
      ADD  SI, SIZE_P3D
      MOV  AX, nextN
      MOV  DS:[SI][0x04], AX
      POP  SI
    }
  else obj->nodes = nextN;
  if (nextN)
    asm {
      PUSH SI
      MOV  SI, nextN
      ADD  SI, SIZE_P3D
      MOV  AX, prevN
      MOV  DS:[SI], AX
      POP  SI
    }
  free(thisN);
  obj->cNodes--;
}

edge *object3D::getEdge(object3Ds *obj, int index) {
  int i;
  edge *lastpnt;
  if (index >= obj->cEdges) return(NULL);
  lastpnt = obj->edges;
  for (i=0;i<index;i++) {
    asm {
      PUSH SI
      MOV  SI, lastpnt
      ADD  SI, SIZE_EDG
      MOV  AX, DS:[SI][0x04]
      MOV  lastpnt, AX
      POP  SI
    }
  }
  lastEdge = lastpnt;
  return(lastpnt);
}

edge *object3D::getNextEdge() {
  edge *next;
  next = lastEdge;
  if (next == NULL) return(NULL);
  asm {
    PUSH SI
    MOV  SI, next
    ADD  SI, SIZE_EDG
    MOV  AX, DS:[SI][0x04]
    MOV  next, AX
    POP  SI
  }
  lastEdge = next;
  return(next);
}

void object3D::addEdge(object3Ds *obj, edge edg) { //don't use with getnext...
  edge *prevN, *nextN;
  edge *newpnt = (edge *)malloc(sizeof(edge)+8); // add space for linked list's pointers
  if (!obj->cEdges) {
    prevN = NULL;
    obj->edges = newpnt;
  } else {
    prevN = getEdge(obj,obj->cEdges-1);
    asm {
      PUSH SI
      MOV  SI, prevN
      ADD  SI, SIZE_EDG
      MOV  AX, newpnt
      MOV  DS:[SI][0x04], AX
      POP  SI
    }
  }
  nextN = NULL;
  asm {
    PUSH SI
    MOV  SI, newpnt
    ADD  SI, SIZE_EDG
    MOV  AX, prevN
    MOV  DS:[SI], AX
    MOV  AX, nextN
    MOV  DS:[SI][0x04], AX
    POP  SI
  }
  newpnt->a = edg.a;
  newpnt->b = edg.b;
  obj->cEdges++;
}

void object3D::delEdge(object3Ds *obj, int index) {
  if (index >= obj->cEdges) return;
  edge *prevE, *thisE, *nextE;
  thisE = getEdge(obj, index);
  asm {
    PUSH SI
    MOV  SI, thisE
    ADD  SI, SIZE_EDG
    MOV  AX, DS:[SI]
    MOV  prevE, AX
    MOV  AX, DS:[SI][0x04]
    MOV  nextE, AX
    POP  SI
  }
  if (prevE)
    asm {
      PUSH SI
      MOV  SI, prevE
      ADD  SI, SIZE_EDG
      MOV  AX, nextE
      MOV  DS:[SI][0x04], AX
      POP  SI
    }
  else obj->edges = nextE;
  if (nextE)
    asm {
      PUSH SI
      MOV  SI, nextE
      ADD  SI, SIZE_EDG
      MOV  AX, prevE
      MOV  DS:[SI], AX
      POP  SI
    }
  free(thisE);
  obj->cEdges--;
}

  /*void object3D::clear(object3Ds *obj) {
    free(obj->name);
    free(obj->nodes);
    free(obj->edges);
  }*/

void object3D::move(object3Ds *obj, point3Ds p3d) {
  int i;
  for (i=0;i<obj->cNodes;i++) {
    getNode(obj, i)->x += p3d.x;
    getNode(obj, i)->y += p3d.y;
    getNode(obj, i)->z += p3d.z;
  }
}

void object3D::scaleO(object3Ds *obj, double scalX, double scalY, double scalZ) {
  int i;
  for (i=0;i<obj->cNodes;i++) {
    getNode(obj, i)->x *= scalX;
    getNode(obj, i)->y *= scalY;
    getNode(obj, i)->z *= scalZ;
  }
}

void object3D::rotateO(object3Ds *obj, double alpha, double beta, double gamma) {
  int i;
  for (i=0;i<obj->cNodes;i++) {
    pointH->rotateO(getNode(obj, i), alpha,beta,gamma);
  }
}

void object3D::rotateP(object3Ds *obj, point3Ds center, double alpha, double beta, double gamma) {
  int i;
  for (i=0;i<obj->cNodes;i++) {
    pointH->rotateP(getNode(obj, i), &center, alpha,beta,gamma);
  }
}

void object3D::draw(object3Ds *obj) {
  int i;
  if (drawMode == 0) {
    pointH->draw(getNode(obj,0));
    for (i=1;i<obj->cNodes;i++) pointH->draw(getNextNode());
  } else {
    point2Ds p1, p2;
    edge *edg;
    int a;
    edg = getEdge(obj,0);
    a = pointH->to2D(getNode(obj,edg->a),&p1);
    a += pointH->to2D(getNode(obj,edg->b),&p2);
    if (a > 1) line(p1.x+160,100-p1.y, p2.x+160,100-p2.y,3);
    for (i=1;i<obj->cEdges;i++) {
      edg = getNextEdge();
      /*a =*/ pointH->to2D(getNode(obj,edg->a),&p1);
      /*a +=*/ pointH->to2D(getNode(obj,edg->b),&p2);
      /*if (a > 1)*/ line(p1.x+160,100-p1.y, p2.x+160,100-p2.y,3);
    }
  }
}

object3Ds *object3D::duplicate(object3Ds *obj) {
  object3Ds *tobj = (object3Ds *)malloc(sizeof(object3Ds));
  init(tobj);
  setName(tobj, obj->name);
  //addNode(tobj, *getNode(obj,0));     // because addnode change lastNode
  for (int i = 0; i<obj->cNodes;i++) {
    //addNode(tobj, *getNextNode());
    addNode(tobj, *getNode(obj,i));
  }
  addEdge(tobj, *getEdge(obj,0));
  for (i = 0; i<obj->cEdges;i++) {
    //addEdge(tobj, *getNextEdge());
    addEdge(tobj, *getEdge(obj,i));
  }
  return(tobj);
}