#ifndef W3D_DECL
  #include <w3d\w3d_decl.h>
#endif
#ifdef W3D_POINT
  #error W3d point already included.
#endif
#define W3D_POINT
#include <math.h>

class point:public pnt_p {       //classe che gestisce i point (2D e 3D)

  private:  //these funcs are used by rotate(point3D)
  double fnX(double x, double y, double theta);
  double fnY(double x, double y, double theta);

  public:
  point();
  void set(point2Ds *p2d, int x, int y);
  void set(point3Ds *p3d, double x, double y, double z);
  point3Ds *to3D(point2Ds *p2d);
  int to2D(point3Ds *p3d, point2Ds *p2d);
  double getDist(point2Ds *p2d1, point2Ds *p2d2);
  double getDist(point3Ds *p3d1, point3Ds *p3d2);
  void rotateO(point2Ds *p2d, double alpha);
  void rotateO(point3Ds *p3d, double alpha, double beta, double gamma);
  void rotateP(point2Ds *p2d, point2Ds *center, double alpha);
  void rotateP(point3Ds *p3d, point3Ds *center, double alpha, double beta, double gamma);
  void draw(point2Ds *p2d, unsigned char color);
  void draw(point3Ds *p3d);
};

point::point() {
  parentObj = NULL;
}

void point::set(point2Ds *p2d, int x, int y) {
  p2d->x = x;
  p2d->y = y;
}

void point::set(point3Ds *p3d, double x, double y, double z) {
  p3d->x = x;
  p3d->y = y;
  p3d->z = z;
}

point3Ds *point::to3D(point2Ds *p2d) {
  // ERRATO! non posso restituire una variabile locale!!
  point3Ds p3dt;
  p3dt.x = p2d->x;
  p3dt.y = p2d->y;
  p3dt.z = 0.0000;
  return(&p3dt);
}

int point::to2D(point3Ds *p3d, point2Ds *p2d) {
  if (parentObj->parentWld->perspective) {
    int sizeZ = parentObj->parentWld->sizeZ;
    int distZ = abs(sizeZ-(parentObj->parentWld->meZ+sizeZ/2));
    if (p3d->z > parentObj->parentWld->meZ) return(0);
    p2d->x = (p3d->x-parentObj->parentWld->meX)*(p3d->z+distZ+sizeZ/2)/sizeZ;
    p2d->y = (p3d->y-parentObj->parentWld->meY)*(p3d->z+distZ+sizeZ/2)/sizeZ;
  } else {
    p2d->x = (int)p3d->x;
    p2d->y = (int)p3d->y;
  }
  return(1);
}

double point::getDist(point2Ds *p2d1, point2Ds *p2d2) {
  return(sqrt(pow(p2d1->x-p2d2->x,2)+pow(p2d1->y-p2d2->y,2)));
}

double point::getDist(point3Ds *p3d1, point3Ds *p3d2) {
  return(sqrt(pow(p3d1->x-p3d2->x,2)+pow(p3d1->y-p3d2->y,2)+pow(p3d1->z-p3d2->z,2)));
}

double point::fnX(double x, double y, double theta) {
  return(x*cos(theta) - y*sin(theta));
}
double point::fnY(double x, double y, double theta) {
  return(y*cos(theta) + x*sin(theta));
}

void point::rotateO(point2Ds *p2d, double alpha) {
  p2d->x = p2d->x*cos(alpha) + p2d->y*sin(alpha);
  p2d->y = p2d->y*cos(alpha) - p2d->x*sin(alpha);
}

void point::rotateO(point3Ds *p3d, double alpha, double beta, double gamma) {
  double tx,ty,tz, rx,ry,rz;
  rx = p3d->x;
  ry = p3d->y;
  rz = p3d->z;
  ty = ry;
  ry = fnX(ty, rz, alpha);
  rz = fnY(ty, rz, alpha);
  tx = rx;
  rx = fnX(tx, rz, beta);
  rz = fnY(tx, rz, beta);
  tx = rx;
  rx = fnX(tx, ry, gamma);
  ry = fnY(tx, ry, gamma);
  p3d->x = rx;
  p3d->y = ry;
  p3d->z = rz;
}

void point::rotateP(point2Ds *p2d, point2Ds *center, double alpha) {
  p2d->x -= center->x;
  p2d->y -= center->y;
  rotateO(p2d, alpha);
  p2d->x += center->x;
  p2d->y += center->y;
}

void point::rotateP(point3Ds *p3d, point3Ds *center, double alpha, double beta, double gamma) {
  p3d->x -= center->x;
  p3d->y -= center->y;
  p3d->z -= center->z;
  rotateO(p3d, alpha, beta, gamma);
  p3d->x += center->x;
  p3d->y += center->y;
  p3d->z += center->z;
}

void point::draw(point2Ds *p2d, unsigned char color) {
  putpixel((unsigned int)p2d->x+160,(unsigned int)100-p2d->y, color);
}

void point::draw(point3Ds *p3d) {
  point2Ds p2d;
  if (to2D(p3d,&p2d)) draw(&p2d,3);
  //int sizeZ = parentObj->parentWld->sizeZ;
  //char cl = (p3d->z+sizeZ/2)*255/sizeZ;*/
}