/*
SNAKE class
*/

#define SNAKE_DIR_E  0
#define SNAKE_DIR_N  1
#define SNAKE_DIR_O  2
#define SNAKE_DIR_S  3

typedef struct {
  short x, y;
} POINT, *PPOINT;

void crt_printMidBlock(short x, short y, BYTE color) {
  char c;
  char oldc = crt_getc(x,y>>1) & 0x00FF;
  BYTE oldatt = crt_getc(x,y>>1) >> 8;
  color &= 0x0F;
  BYTE col = color+16*(oldatt>>4);
  if ((y & 0x01) == 0) c = ''; else c = '';
  if (oldc != ' ') {
    if (oldc != c) col = color+16*(oldatt & 0x0F);
  }
  crt_printc(x, y>>1, col, c);
}

class Snake {
 public:
  Snake();
  ~Snake();
  void setName(char *);
  char *getName() { return(name); }
  int getScore() { return(score); }
  void addScore(int val) { score += val; }
  char getDirection() { return(direction); }
  void setDirection(char dir) { direction = dir; }
  void changeDirection(char);
  BYTE getColor() { return(color); }
  void setColor(BYTE col) { color = col; }
  BYTE getLives() { return(lives); }
  void setLives(BYTE l) { lives = l; }
  short getLength() { return(length); }
  void addPoint(POINT);
  void addPoint(short, short);
  PPOINT getPoint(int);
  void clear();
  void draw();
  void animate();
  char hasCollided();
  char hasCollided(PPOINT);
 private:
  char name[13];
  short length;
  char direction;
  BYTE color, lives;
  int score;
  PERRPROC errFunc;
  POINT *data;

  static int errorHandler(int, int);
};

Snake::Snake() {
  errFunc = errorHandler;
  length = 0;
  direction = SNAKE_DIR_N;
  lives = 3;
  score = 0;
  data = 0;
  color = 0x04;
}

Snake::~Snake() {
  if (data != 0) free(data);
  data = 0;
}

void Snake::setName(char *nam) {
  strncpy(name, nam, 13);
  name[12] = 0;
}

void Snake::changeDirection(char dir) {
  if ((dir & 0x01) == (direction & 0x01)) return;
  direction = dir;
}

void Snake::addPoint(POINT p) {
  if ((length & 0xFF) == 0) data = (PPOINT)realloc(data, sizeof(POINT)*((length & 0xFF00)+0x0100) );
  if (data == 0) errFunc(2, 1);
  *(data+length++) = p;
}

void Snake::addPoint(short x, short y) {
  POINT p;
  p.x = x;
  p.y = y;
  addPoint(p);
}

PPOINT Snake::getPoint(int index) {
  if ((index < 0)||(index >= length)) return(0);
  return( (data+index) );
}

void Snake::clear() {
  if (length > 0) {
    free(data);
    data = 0;
    length = 0;
  }
}

void Snake::draw() {
  for (int i=0; i<length; i++)
    crt_printMidBlock((data+i)->x, (data+i)->y, color);
}

void Snake::animate() {
  POINT newp;
  if (length <= 0) return;
  memmove(data+1, data, sizeof(POINT)*(length-1));
  newp = *(data+1);
  switch (direction) {
    case SNAKE_DIR_N:
      newp.y--;
      break;
    case SNAKE_DIR_S:
      newp.y++;
      break;
    case SNAKE_DIR_O:
      newp.x--;
      break;
    case SNAKE_DIR_E:
      newp.x++;
      break;
  }
  *data = newp;
}

char Snake::hasCollided() { // ret 1 if head hits this snake
  for (int i=1; i<length; i++) {
    if ( (data->x == (data+i)->x)&&(data->y == (data+i)->y) ) return(1);
  }
  return(0);
}

char Snake::hasCollided(PPOINT p) { // returns 2 if head hits p
  for (int i=0; i<length; i++) {
    if ( (p->x == (data+i)->x)&&(p->y == (data+i)->y) ) return((i==0)?2:1);
  }
  return(0);
}

int Snake::errorHandler(int src, int code) {
  textmode(3);
  clrscr();
  printf("Errore %i %i\n", src, code);
  abort();
  return(0); // never executed
}