/*
SNAKEWORLD Class
*/

#define WORLD_MAXSNAKES  5
#define WORLD_MAXDOTS    1

#define WORLD_EMPTY  0
#define WORLD_WALL   1
#define WORLD_DOT1   2   // 5 pts
#define WORLD_DOT2   3   // 20 pts
#define WORLD_DOT3   4   // 10 pts
#define WORLD_DOT4   5   // 30 pts

#define WORLD_COLOR_EMPTY   0+16*1
#define WORLD_COLOR_WALL   12+16*1
#define WORLD_COLOR_DOT1   15+16*1
#define WORLD_COLOR_DOT2    9+16*1
#define WORLD_COLOR_DOT3    7+16*1
#define WORLD_COLOR_DOT4   10+16*1

#define WORLD_SCORE_DOT1    5
#define WORLD_SCORE_DOT2   20
#define WORLD_SCORE_DOT3   10
#define WORLD_SCORE_DOT4   30

int random(int min, int max) { return( rand()%(abs(max-min)+1)+min ); }

class SnakeWorld {
 public:
  SnakeWorld(short, short);
  ~SnakeWorld();
  short getWidth() { return(width); }
  short getHeight() { return(height); }
  void clear();
  void addWall(short, short);
  void addWallLine(short,short, short,short);
  void addSnake(Snake *);
  void draw();
  void animate();
  char getCollision();
  void createDots();
 private:
  short width, height;
  PERRPROC errFunc;
  int n_snakes;
  Snake *snakes[WORLD_MAXSNAKES];
  int n_dots;
  POINT dots[WORLD_MAXDOTS];
  char *data;

  char getData(short, short);
  void setData(short, short, char);
  static int errorHandler(int, int);
};

SnakeWorld::SnakeWorld(short x, short y) {
  width = x;
  height = y;
  n_snakes = 0;
  n_dots = 1;
  errFunc = errorHandler;
  data = (char *)malloc(sizeof(char)*width*height+8);
  if (data == 0) { // error no mem
    errFunc(1, 1);
    return;
  }
  memset(data, WORLD_EMPTY, sizeof(char)*width*height); // empty world
}

SnakeWorld::~SnakeWorld() {
  if (data != 0) free(data);
}

char SnakeWorld::getData(short x, short y) {
  if ((x<0)||(y<0)||(x>=width)||(y>=height)) return(0);
  return(*(data+y*width+x));
}

void SnakeWorld::setData(short x, short y, char d) {
  if ((x<0)||(y<0)||(x>=width)||(y>=height)) return;
  *(data+y*width+x) = d;
}

void SnakeWorld::clear() {
  memset(data, WORLD_EMPTY, sizeof(char)*width*height); // empty world
}

void SnakeWorld::addWall(short x, short y) {
  if ((x<0)||(y<0)||(x>=width)||(y>=height)) return;
  *(data+y*width+x) = WORLD_WALL;
}

void SnakeWorld::addWallLine(short x1,short y1, short x2,short y2) {
  int dx = (x2-x1), dy = (y2-y1);
  int ix = abs(dx), iy = abs(dy);
  int inc = (ix > iy)?ix:iy;
  int px = x1, py = y1;
  int x = 0, y = 0;
  char p;

  addWall(px,py);
  for (int i=0; i<=inc; i++) {
    x += ix;
    y += iy;
    p = 0;
    if (x > inc) {
      p = 1;
      x -= inc;
      if (dx > 0) px++;
      if (dx < 0) px--;
    }
    if (y > inc) {
      p = 1;
      y -= inc;
      if (dy > 0) py++;
      if (dy < 0) py--;
    }
    if (p) addWall(px,py);
  }
}

void SnakeWorld::addSnake(Snake *s) {
  snakes[n_snakes++] = s;
}

void SnakeWorld::draw() {
  BYTE c;
  int i,j;

  for (j=0; j<height; j++)
    for (i=0; i<width; i++) {
      c = getData(i,j);
      switch (c) {
	case WORLD_EMPTY:
	  if ((j & 0x01) == 1) { // down
	    if (getData(i,j-1) == WORLD_EMPTY)
	      crt_printc(i,j >> 1, WORLD_COLOR_EMPTY, ' ');
	    else {
	      short c = crt_getc(i, j>>1);
	      crt_printc(i,j >> 1, WORLD_COLOR_EMPTY+((c >> 8)&0x0F), c & 0xFF);
	    }
	  } else {
	    if (getData(i,j+1) != WORLD_EMPTY) {
	      short c = crt_getc(i, j>>1);
	      crt_printc(i,j >> 1, WORLD_COLOR_EMPTY+((c >> 8)&0x0F), c & 0xFF);
	    }
	  }
	  break;
	case WORLD_WALL: {
	  //crt_printMidBlock(i,j, WORLD_COLOR_WALL);
	  char c;
	  if ((j & 0x01) == 0) { // up
	    c = '';
	  } else {
	    if (getData(i,j-1) == WORLD_WALL) c = ''; else c = '';
	  }
	  crt_printc(i,j>>1, WORLD_COLOR_WALL, c);
	  break;
	}
	case WORLD_DOT1:
	  crt_printMidBlock(i,j, WORLD_COLOR_DOT1);
	  break;
	case WORLD_DOT2:
	  crt_printMidBlock(i,j, WORLD_COLOR_DOT2);
	  break;
	case WORLD_DOT3:
	  crt_printMidBlock(i,j, WORLD_COLOR_DOT3);
	  break;
	case WORLD_DOT4:
	  crt_printMidBlock(i,j, WORLD_COLOR_DOT4);
	  break;
        default:
          crt_printMidBlock(i,j, 0);
          break;
      }
    }
  for (i=0; i<n_snakes; i++) snakes[i]->draw();
}

void SnakeWorld::animate() {
  for (int i=0; i<n_snakes; i++) {
    snakes[i]->animate();
  }
}

char SnakeWorld::getCollision() { // ret -1 for no collision else low nibble is the dead snake
  int i,j;
  PPOINT p;
  for (i=0; i<n_snakes; i++) {
    if (snakes[i]->hasCollided()) return(i);
    p = snakes[i]->getPoint(0); // retrieve head
    if (getData(p->x,p->y) == WORLD_WALL) return(i);
    for (j=0; j<n_snakes; j++) {
      if (j == i) continue;
      char ret = snakes[j]->hasCollided(p);
      if (ret == 1) return(i); else
	if (ret == 2) return(((j << 4)+1)|i);
    }
    for (j=0; j<n_dots; j++) {
      if ((p->x == dots[j].x)&&(p->y == dots[j].y)) {
	BYTE t;
	if (enableSound) {
	  play(300, 50); delay(10);
	  play(500, 40);
	}
	t = getData(p->x, p->y);
        switch (t) {
          case WORLD_DOT1:
            t = WORLD_SCORE_DOT1*multiplier;
            break;
          case WORLD_DOT2:
            t = WORLD_SCORE_DOT2*multiplier;
            break;
          case WORLD_DOT3:
            t = WORLD_SCORE_DOT3*multiplier;
  	    snakeDelay >>= 1;
	    movesLeft = 500;
	    multiplier++;
            break;
        case WORLD_DOT4:
          t = WORLD_SCORE_DOT4*multiplier;
          snakes[i]->setLives(3);
          break;
        }
	snakes[i]->addScore(t);
	snakes[i]->addPoint(*p);
        snakes[i]->addPoint(*p);
	setData(p->x, p->y, WORLD_EMPTY);
	createDots();
      }
    }
  }
  return(-1);
}

void SnakeWorld::createDots() {
  char ok;
  int r,t;
  POINT p;
  for (int i=0; i<n_dots; i++) {
    do {
      ok = 1;
      p.x = random(1,width-2);
      p.y = random(1,height-3);
      if (getData(p.x,p.y) == WORLD_WALL) {
	ok = 0;
	continue;
      }
      for (int j=0; j<n_snakes; j++) {
	if (snakes[j]->hasCollided(&p) != 0) {
	  ok = 0;
	  break;
	}
      }
    } while (!ok);
    dots[i] = p;
    t = WORLD_DOT1;
    r = random(0,5);
    if (r == 1) t = WORLD_DOT2;
    if (r == 2) {
      r = random(0,2);
      if (r == 1) t = WORLD_DOT3;
      if (r == 2) {
        r = random(0,2);
        if (r == 1) t = WORLD_DOT4;
      }
    }
    setData(p.x, p.y, t);
  }
}

int SnakeWorld::errorHandler(int src, int code) {
  textmode(3);
  clrscr();
  printf("Errore %i %i\n", src, code);
  abort();
  return(0); // never executed
}