/*
  Legge da una tabella ed esegue sostituzioni
  
  Author: Daniele Ferro
  Last change: 15/11/2007
*/
// remember to use compact model
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifdef __GNUC__
  #include <glob.h>
#elif defined __MSDOS__
  #include <dos.h>
  #include <dir.h>
#endif

#define MAX_SUBST    3000

const char *default_ext = "sos";

int list_files_count = 0;
char **list_files;

int tab_find_count = 0;
char *tab_find[MAX_SUBST];
char *tab_sost[MAX_SUBST];

/* default options */
char flag_autonames = 1;
char flag_searchmode = 0;
char flag_tabisfile = 0;
char *sost_file, *out_file;

/* ------------------------ */
#ifdef DEBUG
void dbg_print_tab() {
  int i;
  char *empty = "";
  for (i=0; i<tab_find_count; i++) {
    printf("SUBST: %s => %s\n", tab_find[i], flag_searchmode?empty:tab_sost[i]);
  }
}

void dbg_print_files() {
  int i;
  for (i=0; i<list_files_count; i++) {
    printf("FILE: %s\n", list_files[i]);
  }
}
#endif
/* ------------------------ */

void error(int id, char *p) {
  printf("Errore. ");
  switch (id) {
    case 0: puts("Nessun errore"); break;
    case 1: puts("Sintassi errata"); break;
    case 2: puts("Nessun file trovato"); break;
    case 3: puts("Tabella delle sostituzioni non trovata"); break;
    case 4: puts("Tabella delle sostituzioni non valida"); break;
    case 5: printf("Impossibile aprire il file %s\n", p); break;
    case 6: printf("Il file %s esiste gia'\n", p); break;
    case 7: puts("Memoria esaurita"); break;
  }
}

void print_usage() {
  puts("SOS  1.2 - Sostituisce stringhe all'interno di uno o piu' files\n\n"
    "SINTASSI:\n"
    "  SOS [opzioni] inputfile\n"
    "    inputfile e' il file in cui eseguire le ricerche o sostituzioni;\n"
    "    si possono usare caratteri jolly per specificare piu' files\n"
    "   opzioni:\n"
    "     -t tfile    usa la tabella tfile.\n"
    "     -o ofile    usa come nome del file di output ofile, altrimenti viene\n"
    "                 usata la nomenclatura automatica; l'opzione e' valida solo\n"
    "                 se il file di input e' uno solo\n"
    "     -r          esegue solo la ricerca, non esegue sostituzioni\n");
  exit(0);
}

char *del_cr(char *s) {
  int l = strlen(s);
  while (s[l-1] == '\n' || s[l-1] == '\r') l--;
  s[l] = '\0';
  return s;
}

void parse_args(int argc, char *argv[]) {
  int i;
  char *list;
  #ifdef __GNUC__
  glob_t gbuf;
  #elif defined __MSDOS__
  struct ffblk gbuf;
  char drive[MAXDRIVE+MAXDIR+MAXFILE+MAXEXT+8], dir[MAXDIR], name[MAXFILE], ext[MAXEXT],
       ts[MAXDRIVE+MAXDIR+MAXFILE+MAXEXT+8];
  #endif

  if (argc == 1) print_usage();
  if (argc < 2) {
    error(1, NULL);
    print_usage();
  }
  /* reads list of input files */
  list = argv[argc-1];
  #ifdef __GNUC__
  glob(list, 0, NULL, &gbuf);
  list_files = gbuf.gl_pathv;
  if ((list_files_count = gbuf.gl_pathc) == 0) {
    error(2, NULL);
  }
  #elif defined __MSDOS__
  i = findfirst(list, &gbuf, 0);
  list_files = NULL;
  fnsplit(list, drive, dir, name, ext);
  strcat(drive, dir);
  while (!i) {
    list_files_count++;
    list_files = (char**)realloc(list_files, sizeof(char*)*list_files_count);
    if (list_files == NULL) {
      error(7, NULL);
      exit(1);
    }
    strcpy(ts, drive);
    strcat(ts, gbuf.ff_name);
    list_files[list_files_count-1] = strdup(ts);
    i = findnext(&gbuf);
  }
  if (list_files_count == 0) error(2, NULL);
  #endif
  /* reads options */
  for (i=1; i<argc-1; i++) {
    if (argv[i][0] != '-') {
      error(1, NULL);
      print_usage();
    }
    switch (argv[i][1]) {
      case 't':
        if (i == argc-1) { error(1, NULL); print_usage(); }
        sost_file = argv[++i];
        flag_tabisfile = 1;
        break;
      case 'r': flag_searchmode = 1; break;
      case 'o':
        if (i == argc-1 || list_files_count > 1) { error(1, NULL); print_usage(); }
        out_file = argv[++i];
        flag_autonames = 0;
        break;
      default:
        error(1, NULL);
        print_usage();
    }
  }
}

void read_sost() {
  char buffer[1024];
  int k = 0;
  if (flag_tabisfile) {
    FILE *f = fopen(sost_file, "r");
    if (f == NULL) {
      error(3, NULL);
      exit(1);
    }
    while (!feof(f)) {
      buffer[0] = 0;
      fgets(buffer, 1023, f);
      buffer[1023] = '\0';
      del_cr(buffer);
      if (k & 1) {
        free(tab_sost[tab_find_count-1]);
        tab_sost[tab_find_count-1] = strdup(buffer);
      } else {
        if (buffer[0] == 0 || buffer[0] == '\n') { fgets(buffer, 1023, f); continue; }
        tab_find[tab_find_count] = strdup(buffer);
        tab_sost[tab_find_count++] = strdup("");
      }
      k++;
    }
    fclose(f);
  } else {
    if (flag_searchmode) {
      printf("Inserisci le stringhe da cercare, una per riga (termina con una riga vuota):\n");
      do {
        fgets(buffer, 1023, stdin);
        del_cr(buffer);
        if (buffer[0] == '\0') break;
        tab_find[tab_find_count++] = strdup(buffer);
      } while (1);
    } else {
      printf("Inserisci le stringhe da cercare e le relative sostituzioni (termina con una riga vuota):\n");
      do {
        fgets(buffer, 1023, stdin);
        del_cr(buffer);
        if (buffer[0] == '\0') break;
        tab_find[tab_find_count] = strdup(buffer);
        fgets(buffer, 1023, stdin);
        del_cr(buffer);
        tab_sost[tab_find_count++] = strdup(buffer);
      } while (1);
    }
  }
}

char *str_replace(char *line, int *ns) {
  char *nl = strdup(line);
  int i, dp;
  *ns = 0;
  for (i=0; i<tab_find_count; i++) {
    char *lastp;
    dp = 0;
    do {
      char *f;
      int df, nl_len, sost_len, find_len;
      lastp = nl+dp;
      f = strstr(lastp, tab_find[i]);
      if (f == NULL) break;
      df = f-lastp;
      (*ns)++;
      nl_len = strlen(nl);
      sost_len = strlen(tab_sost[i]);
      find_len = strlen(tab_find[i]);
      nl = (char*)realloc(nl, nl_len+sost_len-find_len+2);
      if (nl == NULL) {
        error(7, NULL);
        exit(1);
      }
      lastp = nl+dp;
      f = lastp+df;
      memmove(f+sost_len, f+find_len, strlen(f)-find_len+1);
      memcpy(f, tab_sost[i], sost_len);
      dp += df+sost_len;
    } while (1);
  }
  return nl;
}


void search() {
  int i, j, k;
  long row, tsubst, filesize;
  FILE *f, *fout;
  char buffer[2048];
  for (i=0; i<list_files_count; i++) {
    char *p, *fname = list_files[i];
    char newname[200];
    char printname[50];
    #ifdef __MSDOS__
    char drive[MAXDRIVE+MAXDIR+MAXFILE+MAXEXT+8], dir[MAXDIR], name[MAXFILE], ext[MAXEXT];
    fnsplit(fname, drive, dir, name, ext);
    strcpy(drive, name);
    strcat(drive, ext);
    strncpy(printname,drive, 49);
    #elif define __GNUC__
    p = strrchr(fname, '/');
    strncpy(printname, p?(p+1):fname, 49);
    #endif
    printname[49] = '\0';
    row = 0;
    tsubst = 0;
    f = fopen(fname, "r");
    fout = NULL;
    if (f == NULL) {
      error(5, fname);
      continue;
    }
    fseek(f, 0, SEEK_END);
    filesize = ftell(f);
    rewind(f);
    if (!flag_searchmode) {
      if (flag_autonames) {
        char *p;
        strncpy(newname, fname, 190);
        p = strrchr(newname, '.');
        if (!p) strcat(newname, ".sos");
        else strcpy(p, ".sos");
      } else
        strncpy(newname, out_file, 190);
      fout = fopen(newname, "r");
      if (fout != NULL) {
        fclose(fout);
        error(6, newname);
        continue;
      }
      fout = fopen(newname, "w");
    }
    do {
      buffer[0] = '\0';
      fgets(buffer, 2047, f);
      buffer[2047] = '\0';
      row++;
      del_cr(buffer);
      // only search
      if (flag_searchmode) {
        for (j=0; j<tab_find_count; j++) {
          char *find = tab_find[j];
          char *lastf = buffer;

          while (lastf != NULL) {
            char t[50];
            lastf = strstr(lastf, find);
            if (lastf == NULL) break; else lastf += 1;
            /* found! */
            k = 48-strlen(find);
            if (k < 0) k = 48;
            strncpy(t, lastf-1, k);
            t[k] = '\0';
            printf("%13s: %5li,%3li (%s)   %s\n", printname, row, (lastf-buffer), find, t);
          }
        }                      // also replace
      } else {
        int nsubst;
        char *newline = str_replace(buffer, &nsubst);
        if ((row & 0xff) == 0 ) {
          printf("%s : %i%%\r", printname, (ftell(f)*100)/filesize);
          fflush(stdout);
        }
        tsubst += nsubst;
        fprintf(fout, "%s\n", newline);
        free(newline);
      }
    } while (!feof(f));
    fclose(f);
    if (fout) fclose(fout);
    if (!flag_searchmode) {
      printf("%13s: %li sostituzioni\n", printname, tsubst);
    }
  }
}

int main(int argc, char *argv[]) {
  /* reads arguments */
  parse_args(argc, argv);
  /* reads sost table */
  read_sost();
  /* do search */
  search();
  return(0);
}
