/* SPACE.H */

typedef double datatype;

typedef struct {
  datatype x, y, z;
} point;

class Projector {
 public:
  char enableBackPlane;

  Projector();
  Projector(point *me, point *core);
  void setMe(point *m);
  void setMe(datatype x, datatype y, datatype z);
  point *getMe() { return( &me ); }
  point *getOrigin() { return( &eye ); }
  void moveMe(point *m);
  void moveMe(datatype x, datatype y, datatype z);
  void moveMeRel(datatype x, datatype y, datatype z);
  void setOrigin(point *o);
  void setOrigin(datatype x, datatype y, datatype z);
  void moveOriginCloser(datatype amount);
  void setUnits(datatype ux, datatype uy) { unitx = ux; unity = uy; }
  datatype calcDistMe(point *p);
  datatype calcDistMe(datatype x, datatype y, datatype z);
  datatype calcDistOrigin(point *p);

  void rotateMe(point *center, datatype alpha, datatype beta, datatype gamma);
  void rotateMeRel(point *center, datatype alpha, datatype beta, datatype gamma);
  int trace(point *p, int *x, int *y);
 private:
  typedef struct {
    datatype x0, x1, x2, x3;
  } proint;

  point me, eye;
  point dir1, dir2;
  datatype unitx, unity;

  proint plane;
  datatype planelen, dirlen1, dirlen2;

  datatype mxRot[3][9];
  datatype alphas[3];

  void calcDir(point *p1, point *p2, point *dir1, point *dir2);
  void calcPlane();
  void calcMxRot(datatype alpha, datatype beta, datatype gamma);
  void calcMxRot(point *axis, datatype alpha, int nmatrix);

  void sup_2p(point *p1, point *p2, proint *sup1, proint *sup2);
  datatype scal_prod(point *p1, point *p2);
  void vect_prod(point *p1, point *p2, point *res);
};

Projector::Projector() {
  me.x = me.y = me.z = 0;
  eye.x = 0;
  eye.y = 0;
  eye.z = -100;
  unitx = unity = 100;
  enableBackPlane = 1;
  calcPlane();
}

Projector::Projector(point *me, point *core) {
  this->me = *me;
  this->eye = *core;
  unitx = unity = 100;
  enableBackPlane = 1;
  calcPlane();
}

inline void Projector::setMe(point *m) {
  me = *m;
  calcPlane();
}

inline void Projector::setMe(datatype x, datatype y, datatype z) {
  me.x = x; me.y = y; me.z = z;
  calcPlane();
}

inline void Projector::moveMe(point *m) {
  me.x += m->x;
  me.y += m->y;
  me.z += m->z;
  eye.x += m->x;
  eye.y += m->y;
  eye.z += m->z;
  // I should change also v1.x0 and v2.x0 but they're never used
  plane.x0 = -(me.x*plane.x1+me.y*plane.x2+me.z*plane.x3);
}

inline void Projector::moveMe(datatype x, datatype y, datatype z) {
  me.x += x;
  me.y += y;
  me.z += z;
  eye.x += x;
  eye.y += y;
  eye.z += z;
  plane.x0 = -(me.x*plane.x1+me.y*plane.x2+me.z*plane.x3);
}

void Projector::moveMeRel(datatype x, datatype y, datatype z) {
  point delta;
  delta.x = z*plane.x1/planelen - x*dir1.x/dirlen1 - y*dir2.x/dirlen2;
  delta.y = z*plane.x2/planelen - x*dir1.y/dirlen1 - y*dir2.y/dirlen2;
  delta.z = z*plane.x3/planelen - x*dir1.z/dirlen1 - y*dir2.z/dirlen2;

  me.x += delta.x;
  me.y += delta.y;
  me.z += delta.z;
  eye.x += delta.x;
  eye.y += delta.y;
  eye.z += delta.z;
  plane.x0 = -(me.x*plane.x1+me.y*plane.x2+me.z*plane.x3);
}

inline void Projector::setOrigin(point *o) {
  eye = *o;
  calcPlane();
}

inline void Projector::setOrigin(datatype x, datatype y, datatype z) {
  eye.x = x; eye.y = y; eye.z = z;
  calcPlane();
}

void Projector::moveOriginCloser(datatype amount) {
  double t = amount/planelen;
  eye.x += plane.x1*t;
  eye.y += plane.x2*t;
  eye.z += plane.x3*t;
  plane.x1 = me.x-eye.x;
  plane.x2 = me.y-eye.y;
  plane.x3 = me.z-eye.z;
  plane.x0 = -(me.x*plane.x1+me.y*plane.x2+me.z*plane.x3);
  planelen = sqrt(plane.x1*plane.x1+plane.x2*plane.x2+plane.x3*plane.x3);
}

datatype Projector::calcDistMe(point *p) {
  return( sqrt((me.x-p->x)*(me.x-p->x) + (me.y-p->y)*(me.y-p->y) + (me.z-p->z)*(me.z-p->z)) );
}

datatype Projector::calcDistMe(datatype x, datatype y, datatype z) {
  return( sqrt((me.x-x)*(me.x-x) + (me.y-y)*(me.y-y) + (me.z-z)*(me.z-z)) );
}

datatype Projector::calcDistOrigin(point *p) {
  return( sqrt((eye.x-p->x)*(eye.x-p->x) + (eye.y-p->y)*(eye.y-p->y) + (eye.z-p->z)*(eye.z-p->z)) );
}

void Projector::calcMxRot(datatype alpha, datatype beta, datatype gamma) {
  double s[3], c[3]; // sine/cosine tables
  static olda = 0, oldb = 0, oldc = 0;
  if (alpha == olda && beta == oldb && gamma == oldc) return;
  olda = alpha;
  oldb = beta;
  oldc = gamma;
  /*s[0] = sin(alpha);
  c[0] = cos(alpha);
  s[1] = sin(beta);
  c[1] = cos(beta);
  s[2] = sin(gamma);
  c[2] = cos(gamma);*/
  asm {
    fld   alpha
    dw 0xFBD9 // fsincos
    fstp  qword ptr [c+0]
    fstp  qword ptr [s+0]
    fld   beta
    dw 0xFBD9 // fsincos
    fstp  qword ptr [c+8]
    fstp  qword ptr [s+8]
    fld   gamma
    dw 0xFBD9 // fsincos
    fstp  qword ptr [c+16]
    fstp  qword ptr [s+16]
  }
  mxRot[0][0] = c[1]*c[2];
  mxRot[0][1] = -(s[0]*s[1]*c[2]+c[0]*s[2]);
  mxRot[0][2] = s[0]*s[2]-c[0]*s[1]*c[2];
  mxRot[0][3] = c[1]*s[2];
  mxRot[0][4] = c[0]*c[2]-s[0]*s[1]*s[2];
  mxRot[0][5] = -(c[0]*s[1]*s[2]+s[0]*c[2]);
  mxRot[0][6] = s[1];
  mxRot[0][7] = s[0]*c[1];
  mxRot[0][8] = c[0]*c[1];
}

void Projector::rotateMe(point *center, datatype alpha, datatype beta, datatype gamma) {
  double rx,ry,rz;

  calcMxRot(alpha, beta, gamma);

  if (me.x != center->x || me.y != center->y || me.z != center->z) {
    rx = me.x - center->x;
    ry = me.y - center->y;
    rz = me.z - center->z;
    me.x = rx*mxRot[0][0] + ry*mxRot[0][1] + rz*mxRot[0][2] + center->x;
    me.y = rx*mxRot[0][3] + ry*mxRot[0][4] + rz*mxRot[0][5] + center->y;
    me.z = rx*mxRot[0][6] + ry*mxRot[0][7] + rz*mxRot[0][8] + center->z;
  }

  if (eye.x != center->x || eye.y != center->y || eye.z != center->z) {
    rx = eye.x - center->x;
    ry = eye.y - center->y;
    rz = eye.z - center->z;
    eye.x = rx*mxRot[0][0] + ry*mxRot[0][1] + rz*mxRot[0][2] + center->x;
    eye.y = rx*mxRot[0][3] + ry*mxRot[0][4] + rz*mxRot[0][5] + center->y;
    eye.z = rx*mxRot[0][6] + ry*mxRot[0][7] + rz*mxRot[0][8] + center->z;
  }

  rx = dir1.x;
  ry = dir1.y;
  rz = dir1.z;
  dir1.x = rx*mxRot[0][0] + ry*mxRot[0][1] + rz*mxRot[0][2];
  dir1.y = rx*mxRot[0][3] + ry*mxRot[0][4] + rz*mxRot[0][5];
  dir1.z = rx*mxRot[0][6] + ry*mxRot[0][7] + rz*mxRot[0][8];

  rx = dir2.x;
  ry = dir2.y;
  rz = dir2.z;
  dir2.x = rx*mxRot[0][0] + ry*mxRot[0][1] + rz*mxRot[0][2];
  dir2.y = rx*mxRot[0][3] + ry*mxRot[0][4] + rz*mxRot[0][5];
  dir2.z = rx*mxRot[0][6] + ry*mxRot[0][7] + rz*mxRot[0][8];

  plane.x1 = me.x-eye.x;
  plane.x2 = me.y-eye.y;
  plane.x3 = me.z-eye.z;
  plane.x0 = -(me.x*plane.x1+me.y*plane.x2+me.z*plane.x3);

}

void Projector::calcMxRot(point *axis, datatype alpha, int nmatrix) {
  double s = sin(alpha), c = cos(alpha);
  if (alphas[nmatrix] == alpha) return;
  alphas[nmatrix] = alpha;
  mxRot[nmatrix][0] = c + (1-c)*axis->x*axis->x;
  mxRot[nmatrix][1] = (1-c)*axis->x*axis->y - s*axis->z;
  mxRot[nmatrix][2] = (1-c)*axis->x*axis->z + s*axis->y;
  mxRot[nmatrix][3] = (1-c)*axis->y*axis->x + s*axis->z;
  mxRot[nmatrix][4] = c + (1-c)*axis->y*axis->y;
  mxRot[nmatrix][5] = (1-c)*axis->y*axis->z - s*axis->x;
  mxRot[nmatrix][6] = (1-c)*axis->z*axis->x - s*axis->y;
  mxRot[nmatrix][7] = (1-c)*axis->z*axis->y + s*axis->x;
  mxRot[nmatrix][8] = c + (1-c)*axis->z*axis->z;
}

void Projector::rotateMeRel(point *center, datatype alpha, datatype beta, datatype gamma) {
  double rx,ry,rz;
  /*point oeye = eye;
  point odir1 = dir2;
  point odir2 = dir2;*/

  /*gotoxy(1,1);
  printf("dir1-2: %lf\n", sqrt((dir1.x-dir2.x)*(dir1.x-dir2.x)+(dir1.y-dir2.y)*(dir1.y-dir2.y)+(dir1.z-dir2.z)*(dir1.z-dir2.z)));
  printf("eye-dir1: %lf\n", sqrt((dir1.x-eye.x+me.x)*(dir1.x-eye.x+me.x)+(dir1.y-eye.y+me.y)*(dir1.y-eye.y+me.y)+(dir1.z-eye.z+me.z)*(dir1.z-eye.z+me.z)));
  printf("eye-dir2: %lf\n", sqrt((dir2.x-eye.x+me.x)*(dir2.x-eye.x+me.x)+(dir2.y-eye.y+me.y)*(dir2.y-eye.y+me.y)+(dir2.z-eye.z+me.z)*(dir2.z-eye.z+me.z)));
  */
  if (alpha != 0) {
    calcMxRot(&dir1, alpha, 0);
    rx = (eye.x - me.x);
    ry = (eye.y - me.y);
    rz = (eye.z - me.z);
    eye.x = rx*mxRot[0][0] + ry*mxRot[0][1] + rz*mxRot[0][2] + me.x;
    eye.y = rx*mxRot[0][3] + ry*mxRot[0][4] + rz*mxRot[0][5] + me.y;
    eye.z = rx*mxRot[0][6] + ry*mxRot[0][7] + rz*mxRot[0][8] + me.z;
    rx = dir2.x;
    ry = dir2.y;
    rz = dir2.z;
    dir2.x = rx*mxRot[0][0] + ry*mxRot[0][1] + rz*mxRot[0][2];
    dir2.y = rx*mxRot[0][3] + ry*mxRot[0][4] + rz*mxRot[0][5];
    dir2.z = rx*mxRot[0][6] + ry*mxRot[0][7] + rz*mxRot[0][8];
  }

  if (beta != 0) {
    calcMxRot(&dir2, beta, 1);
    rx = eye.x - me.x;
    ry = eye.y - me.y;
    rz = eye.z - me.z;
    eye.x = rx*mxRot[1][0] + ry*mxRot[1][1] + rz*mxRot[1][2] + me.x;
    eye.y = rx*mxRot[1][3] + ry*mxRot[1][4] + rz*mxRot[1][5] + me.y;
    eye.z = rx*mxRot[1][6] + ry*mxRot[1][7] + rz*mxRot[1][8] + me.z;
    rx = dir1.x;
    ry = dir1.y;
    rz = dir1.z;
    dir1.x = rx*mxRot[1][0] + ry*mxRot[1][1] + rz*mxRot[1][2];
    dir1.y = rx*mxRot[1][3] + ry*mxRot[1][4] + rz*mxRot[1][5];
    dir1.z = rx*mxRot[1][6] + ry*mxRot[1][7] + rz*mxRot[1][8];
  }

  /* // NOT IMPLEMENTED YET
  if (gamma != 0) {
    point v = {me.x-oeye.x, me.y-oeye.y, me.z-oeye.z};
    calcMxRot(&v, gamma, 2);
    rx = dir1.x;
    ry = dir1.y;
    rz = dir1.z;
    dir1.x = rx*mxRot[2][0] + ry*mxRot[2][1] + rz*mxRot[2][2];
    dir1.y = rx*mxRot[2][3] + ry*mxRot[2][4] + rz*mxRot[2][5];
    dir1.z = rx*mxRot[2][6] + ry*mxRot[2][7] + rz*mxRot[2][8];
    rx = dir2.x;
    ry = dir2.y;
    rz = dir2.z;
    dir2.x = rx*mxRot[2][0] + ry*mxRot[2][1] + rz*mxRot[2][2];
    dir2.y = rx*mxRot[2][3] + ry*mxRot[2][4] + rz*mxRot[2][5];
    dir2.z = rx*mxRot[2][6] + ry*mxRot[2][7] + rz*mxRot[2][8];
  }
  */

  plane.x1 = me.x-eye.x;
  plane.x2 = me.y-eye.y;
  plane.x3 = me.z-eye.z;
  plane.x0 = -(me.x*plane.x1+me.y*plane.x2+me.z*plane.x3);
}


datatype Projector::scal_prod(point *p1, point *p2) {
  return p1->x*p2->x+p1->y*p2->y+p1->z*p2->z;
}

void Projector::vect_prod(point *p1, point *p2, point *res) {
  res->x = p1->y*p2->z-p1->z*p2->y;
  res->y = p1->z*p2->x-p1->x*p2->z;
  res->z = p1->x*p2->y-p1->y*p2->x;
}

void Projector::calcDir(point *p1, point *p2, point *dir1, point *dir2) {
  datatype l, m, n;
  l = p2->x-p1->x;
  m = p2->y-p1->y;
  n = p2->z-p1->z;

  dir1->x = dir1->y = dir1->z = 0;
  dir2->x = dir2->y = dir2->z = 0;
  if (n != 0) {
    /*dir1->x = p2->z-p1->z;
    dir1->z = p1->x-p2->x;
    dir2->y = p2->z-p1->z;
    dir2->z = p1->y-p2->y;*/
    dir1->x = n;
    dir1->z = -l;
    dir2->y = n;
    dir2->z = -m;
  } else if (m != 0) {
    /*dir1->x = p2->y-p1->y;
    dir1->y = p1->x-p2->x;
    dir2->y = p2->z-p1->z;
    dir2->z = p1->y-p2->y;*/
    dir1->x = m;
    dir1->y = -l;
    dir2->y = n;
    dir2->z = -m;
  } else {
    /*dir1->x = p2->y-p1->y;
    dir1->y = p1->x-p2->x;
    dir2->x = p2->z-p1->z;
    dir2->z = p1->x-p2->x;*/
    dir1->x = m;
    dir1->y = -l;
    dir2->x = n;
    dir2->z = -l;
  }
}

void Projector::calcPlane() {
  // direzioni
  calcDir(&me, &eye, &dir1, &dir2);

  // piano perpendicolare a r passante per me
  plane.x1 = me.x-eye.x;
  plane.x2 = me.y-eye.y;
  plane.x3 = me.z-eye.z;
  plane.x0 = -(me.x*plane.x1+me.y*plane.x2+me.z*plane.x3);
  planelen = sqrt(plane.x1*plane.x1+plane.x2*plane.x2+plane.x3*plane.x3);

  // lunghezze versori
  dirlen1 = sqrt(dir2.x*dir2.x+dir2.y*dir2.y+dir2.z*dir2.z);
  dirlen2 = sqrt(dir1.x*dir1.x+dir1.y*dir1.y+dir1.z*dir1.z);
}

int Projector::trace(point *p, int *posx, int *posy) {
  point pro;

  //gotoxy(1,1);

  //printf("me: %.3lf %.3lf %.3lf\n", me.x, me.y, me.z);
  //printf("retta oe: (%+07.3lf %+07.3lf %+.3lf %+07.3lf)(%+07.3lf %+.3lf %+.3lf %+.3lf)\n", v1.x0, v1.x1,v1.x2,v1.x3, v2.x0,v2.x1,v2.x2,v2.x3);
  //printf("piano: (%+09.3lf %+07.3lf %+07.3lf %+07.3lf)\n", plane.x0, plane.x1,plane.x2,plane.x3);

  // inters piano e retta eye-p
  datatype tp = (plane.x0+plane.x1*eye.x+plane.x2*eye.y+plane.x3*eye.z);
  if (tp == 0) return(-1);
  datatype t = (plane.x0+plane.x1*p->x+plane.x2*p->y+plane.x3*p->z)/tp;
  if (t == 1) return(-2);
  pro.x = (p->x - eye.x*t)/(1-t);
  pro.y = (p->y - eye.y*t)/(1-t);
  pro.z = (p->z - eye.z*t)/(1-t);

  //printf("pro: (%.3lf %.3lf %.3lf %.3lf)\n", pro.x0, pro.x1, pro.x2, pro.x3);

  point i1, t1, t2;
  datatype x,y, a,b;

  i1.x = pro.x-me.x; i1.y = pro.y-me.y; i1.z = pro.z-me.z;

  /*printf("v1: (%.3lf, %.3lf, %.3lf, %.3lf)\n", v1.x0,v1.x1,v1.x2,v1.x3);
  printf("v2: (%.3lf, %.3lf, %.3lf, %.3lf)\n", v2.x0,v2.x1,v2.x2,v2.x3);*/

  // controlla se il punto  dietro a plane
  point i2 = {p->x-eye.x, p->y-eye.y, p->y-eye.y};
  if (i2.x != 0) {
    a = (pro.x-eye.x)/i2.x;
  } else if (i2.y != 0) {
    a = (pro.y-eye.y)/i2.y;
  } else {
    a = (pro.z-eye.z)/i2.z;
  }
  if (a <= 0) return(-3); // skip if behind eye
  if (!enableBackPlane)
    if (1 <= a) return(-4); // skip if behind plane

  x = -scal_prod(&i1, &dir1)/dirlen1;
  y = scal_prod(&i1, &dir2)/dirlen2;

  // scala x,y
  *posx = 320*(x+unitx)/(unitx*2);
  *posy = 200*(y+unity)/(unity*2);
  return(0);
}


//-------------------------------------------------------------------

/*void Projector::sup_2p(point *p1, point *p2, proint *sup1, proint *sup2) {
  datatype l, m, n;
  l = p2->x-p1->x;
  m = p2->y-p1->y;
  n = p2->z-p1->z;

  if ((l==0 && m == 0)||(n != 0)) {
    sup1->x0 = p1->z*p2->x-p1->x*p2->z;
    sup1->x1 = p2->z-p1->z;
    sup1->x2 = 0;
    sup1->x3 = p1->x-p2->x;
    sup2->x0 = p1->z*p2->y-p1->y*p2->z;
    sup2->x1 = 0;
    sup2->x2 = p2->z-p1->z;
    sup2->x3 = p1->y-p2->y;
  } else if ((l==0 && n == 0)||(m != 0)) {
    sup1->x0 = p2->x*p1->y-p2->y*p1->x;
    sup1->x1 = p2->y-p1->y;
    sup1->x2 = p1->x-p2->x;
    sup1->x3 = 0;
    sup2->x0 = p2->y*p1->z-p2->z*p1->y;
    sup2->x1 = 0;
    sup2->x2 = p2->z-p1->z;
    sup2->x3 = p1->y-p2->y;
  } else {
    sup1->x0 = p1->x*p2->y-p1->y*p2->x;
    sup1->x1 = p2->y-p1->y;
    sup1->x2 = p1->x-p2->x;
    sup1->x3 = 0;
    sup2->x0 = p1->x*p2->z-p1->z*p2->x;
    sup2->x1 = p2->z-p1->z;
    sup2->x2 = 0;
    sup2->x3 = p1->x-p2->x;
  }

  if (sup1->x0 != 0) {
    sup1->x1 /= sup1->x0;
    sup1->x2 /= sup1->x0;
    sup1->x3 /= sup1->x0;
  }
  if (sup2->x0 != 0) {
    sup2->x1 /= sup2->x0;
    sup2->x2 /= sup2->x0;
    sup2->x3 /= sup2->x0;
  }
}*/
