#pragma option -2
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <dos.h>

#include <math.h>

#include <dtypes.h>
#include <crt.h>
#include <vga.h>
#include <vgapal.h>

const RGBcol black = {0,0,0},
	     white = {63,63,63};

// special functions for packed-colors
void expandCol(BYTE color, RGBcol *colOut) {
  asm {
    push  si
    mov	  si, colOut
    mov	  dl, color
    mov   ax, dx
    and	  dx, 0x00C0
    shr   dl, 6
    inc   dl
    mov   ah, dl  // |
    shl   dl, 2   // |
    add   dl, ah  // | dl = dl * 5
    mov   ah, al
    xor   bx, bx
    mov   bl, al
    and   ax, 0x0C03
    and   bx, 0x0030
    shr   ah, 2
    shr   bl, 4
    db 0x0F,0xAF,0xC2     // imul  ax, dx
    db 0x0F,0xAF,0xDA     // imul  bx, dx
    mov   [si], ax
    mov   [si+2], bx
    pop   si
  }
  /*BYTE t = ((color & 0xC0) >> 6)+1;            // 1..4
  colOut->b = (color & 0x03)*t*5;              // 0..60
  colOut->g = ((color & 0x0C) >> 2)*t*5;       // 0..60
  colOut->r = ((color & 0x30) >> 4)*t*5;       // 0..60*/
}

BYTE compactCol(RGBcol color) {
  BYTE t1,t2,t3;
  t1 = min(color.r/20+1,4);
  t2 = min(color.g/20+1,4);
  t3 = min(color.b/20+1,4);
  t1 = max(max(t1,t2), t3);
  register BYTE q = 5*t1;
  color.r = (color.r/q);
  color.g = (color.g/q);
  color.b = (color.b/q);
  asm {
    and   WORD ptr color, 0x0303
    and   WORD ptr color[2], 0x0003
    dec   t1
    shl   t1, 6
  }
  return(t1|(color.b << 4)|(color.g << 2)|color.r);
}

void palSetPacked(pal256 *palette) {
  for (int i=0; i<256; i++) {
    RGBcol color;
    expandCol(i, &color);
    (*palette)[i] = color;
  }
}

class virtualScreenPack: public VirtualScreen {
  public:
    void putPixel(WORD x, WORD y, BYTE color);
    void blur(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2);
};

#pragma option -w-ccc
void virtualScreenPack::putPixel(WORD x, WORD y, BYTE color) {
  //if ((x < VGA_MARGIN)||(y < VGA_MARGIN)||(x > VGA_MAXX-VGA_MARGIN)||(y > VGA_MAXY-VGA_MARGIN)) return;
  BYTE far *scrI = bufferIn+y*VGA_WIDTH+x,
       far *scrO = bufferOut+y*VGA_WIDTH+x;;
  RGBcol c1,c2;
  expandCol(color, &c1);
  expandCol(*scrI, &c2);
  c1.r = min(c1.r+c2.r, 63);
  c1.g = min(c1.g+c2.g, 63);
  c1.b = min(c1.b+c2.b, 63);
  color = compactCol(c1);
  *scrO = color;
}
#pragma option -w+ccc

void virtualScreenPack::blur(VGACOORD x1, VGACOORD y1, VGACOORD x2, VGACOORD y2) {
  RGBcol b1,b2,b3,b4;
  BYTE far *scrI;
  register int i, j;
  if (!checkCoords(&x1,&y1,&x2,&y2)) return;
  for (j=y1;j<y2;j++) {
    for (i=x1;i<=x2;i++) {
      scrI = bufferIn+j*VGA_WIDTH+i;
      expandCol(*scrI, &b1);
      expandCol(*(scrI+VGA_WIDTH-1), &b2);
      expandCol(*(scrI+VGA_WIDTH+1), &b3);
      expandCol(*(scrI+VGA_WIDTH*2), &b4);
      b1.r = (b1.r+b2.r+b3.r+b4.r) >> 2;
      b1.g = (b1.g+b2.g+b3.g+b4.g) >> 2;
      b1.b = (b1.b+b2.b+b3.b+b4.b) >> 2;
      *(bufferOut+(j+1)*VGA_WIDTH+i) = compactCol(b1);
    }
  }
}

//---------------------------------------------------------------------------

void pol2rec_f4(float rho, float theta, float *x, float *y) {
// 387, near model
  __emit__(
    0x56,                  // push  si
    0x9B,0xDB,0xE3,        // finit
    0xD9,0x46,&theta,      // fld   theta
    0xD9,0xFB,             // fsincos
    0xD8,0x4E,&rho,        // fmul  rho
    0x8B,0x76,&x,          // mov   si, x
    0xD9,0x1C,             // fstp  [si]
    0xD8,0x4E,&rho,        // fmul  rho
    0x8B,0x76,&y,          // mov   si, y
    0xD9,0x14,             // fst   [si]
    0x5E                   // pop   si
  );
}

void rec2pol_f4(float x, float y, float *rho, float *theta) {
  __emit__(
    0x56,                  // push  si
    0x9B,0xDB,0xE3,        // finit
    0xD9,0x46,&x,	   // fld   x
    0xD8,0x4E,&x,          // fmul  x
    0xD9,0x46,&y,	   // fld   y
    0xD8,0x4E,&y,          // fmul  y
    0xDE,0xC1,             // faddp st1, st0
    0xD9,0xFA,             // fsqrt
    0x8B,0x76,&rho,        // mov   si, rho
    0xD9,0x14,             // fst   [si]
    0xD9,0x46,&y,	   // fld   y
    0xD9,0x46,&x,	   // fld   x
    0xD9,0xF3,             // fpatan
    0x8B,0x76,&theta,      // mov   si, theta
    0xD9,0x14,             // fst   [si]
    0x5E                   // pop   si
  );
}

void trans(VGACOORD x,VGACOORD y, VGACOORD *xout,VGACOORD *yout) {
  float a,d, xf,yf;
  rec2pol_f4((int)x-160,(int)y-100, &d,&a);
  a = a+0.05;
  d = d*0.9;
  pol2rec_f4(d,a, &xf,&yf);
  *xout = (WORD)xf+160;
  *yout = (WORD)yf+100;

  //*xout = ((int)x-160)/1.2+160;
  //*yout = ((int)y-100)/1.2+100;
}


void main(void) {

  pal256 pal1;
  VirtualScreen *vir1 = new VirtualScreen();
  VirtualScreen *vir2 = new VirtualScreen();
  VirtualScreen *vir3 = new VirtualScreen();

  crt_setVideoMode(0x13);

  pal_setGradient(&pal1, 0,100, 0,0,0, 5,55,40);
  pal_setGradient(&pal1, 101,200, 5,55,40, 10,20,40);
  pal_setGradient(&pal1, 201,255, 10,20,40, 63,63,63);

  //palSetPacked(&pal1);
  pal_outAll(&pal1);

  vir1->clear(0);
  vir2->clear(0);
  vir3->clear(0);
  int i = 0;
  const textSize = 4,
	textX = 40,
	textY = 80,
	zoom = 30,
	ofsX = 20;


  while (!kbhit()) {
    i = ++i % 360;
    vir1->text(textX+(int)(sin(8*i*M_PI/180)*20),textY,"Welcome!",textSize,60);
    vir1->copyTo(vir2);

    for (int k=1;k<4;k++) {
      vir2->blur(0,textY-10,319,textY+16*textSize+10);

      vir2->bufferOut = vir3->bufferIn;
      vir2->stretch(textX+zoom*k,textY, textX+8*8*textSize-zoom*k,textY+16*textSize,
		    0,70-ofsX*k, 319,70+16*textSize-ofsX*k);
      vir2->restoreBuffers();

      vir1->drawMode = DM_ADD;
      vir3->bufferOut = vir1->bufferIn;
      vir3->copyRectTo(vir1, 0,70-ofsX*k,319, 70+16*textSize-ofsX*k);
      vir1->drawMode = DM_NORMAL;
      vir3->restoreBuffers();

      vir3->clear(0);
    }

    vir1->blur(0,0,319,199);
    vir1->show();
    vir1->clear(0);
    delay(10);
  }
  getch();

  vir2->clear(0);
  while (!kbhit()) {
    i = ++i % 360;
    vir2->copyTo(vir1);
    vir1->text(30+(int)(sin(8*i*M_PI/180)*30),70+(int)(cos(8*i*M_PI/180)*50),"Welcome!",textSize,16);

    vir1->drawMode = DM_SUB2;
    vir1->bufferOut = vir2->bufferIn;
    vir1->move(0,0,319,199, trans);
    vir1->restoreBuffers();
    vir1->drawMode = DM_NORMAL;

    vir2->text(30+(int)(sin(8*i*M_PI/180)*30),70+(int)(cos(8*i*M_PI/180)*50),"Welcome!",textSize,250);

    vir2->blur(0,0,319,199);
    vir2->blur(0,0,319,199);
    vir2->show();
    delay(1);
  }
  getch();

  int j = 0;
  vir2->copyTo(vir1);
  pal_inAll(&pal1);
  pal_setGradient(&pal1, 1,4, 0,0,0, 63,63,63);
  pal1[0] = black;
  for (i=1;i<18;i++) {
    if (i> 14) j += 2;
    vir2->clear(0);
    pal_light(&pal1, 5,255, 3);
    pal_outAll(&pal1);
    vir1->bufferOut = vir2->bufferIn;
    vir1->stretch(0,0,319,199, 0+j,0+i*2,319-j,199-i*2);
    vir1->restoreBuffers();
    vir2->copyTo(vir1);
    vir1->blur(0,0+i*2,319,199-i*2);
    vir1->blur(0,0+i*2,319,199-i*2);
    vir1->show();
    delay(40);
  }
  for (i=1;i<22;i++) {
    vir2->clear(0);
    pal_light(&pal1, 5,255, 3);
    pal_outAll(&pal1);
    vir1->bufferOut = vir2->bufferIn;
    vir1->stretch(0,80,319,199-80, 0+i*2,80,319-i*2,199-82);
    vir1->restoreBuffers();
    vir2->copyTo(vir1);
    vir1->show();
    delay(50);
  }
  for (i=0;i<10;i++) {
    pal_dark(&pal1, 0,255, 2);
    pal_outAll(&pal1);
    vir1->blur(0+100,0+80,319-100,199-80);
    vir1->show();
    delay(100);
  }

  delete vir3;
  delete vir2;
  delete vir1;
  crt_setVideoMode(0x03);
}
